"""
This file contains the classes to perform rule classification evaluation.

"""
import numpy as np
from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score, confusion_matrix

try:
    from . import rules
    from . import fuzzy_sets as fs
except ImportError:
    import rules
    import fuzzy_sets as fs


class evalRuleBase():
    '''
    Class to evaluate a set of rules given a evaluation dataset.
    '''

    def __init__(self, mrule_base: rules.MasterRuleBase, X: np.array, y: np.array, time_moments: np.array=None, precomputed_truth=None) -> None:
        '''
        Creates the object with the rulebase to evaluate and the data to use in the evaluation.

        :param mrule_base: The rule base to evaluate.
        :param X: array shape samples x features. The data to evaluate the rule base.
        :param y: array shape samples x 1. The labels of the data.
        :param time_moments: array shape samples x 1. The time moments of the samples. (Only for temporal rule bases)
        :return: None
        '''
        self.mrule_base = mrule_base
        self.X = X
        self.y = y
        self.time_moments = time_moments

        self.consequents = mrule_base.get_consequents()
        self.consequents_names = mrule_base.get_consequents_names()
        self.precomputed_truth = precomputed_truth

        if isinstance(y[0], str):
            self.y = np.array([list(self.consequents_names).index(str(y)) for y in y])



    def compute_pattern_support(self) -> np.array:
        '''
        Computes the pattern support for each of the rules for the given X.
        Each pattern support firing strength is the result of the tnorm for all the antecedent memeberships,
        dvided by their number.

        :return: array of shape rules x 2
        '''
        if self.time_moments is None:
            antecedent_memberships = self.mrule_base.compute_firing_strenghts(
                self.X, precomputed_truth=self.precomputed_truth)
        else:
            antecedent_memberships = self.mrule_base.compute_firing_strenghts(
                self.X, self.time_moments)

        patterns = self._get_all_rules()

        if self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.t1:
            res = np.zeros((len(patterns), ))
        elif self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.t2:
            res = np.zeros((len(patterns), 2))
        elif self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.gt2:
            res = np.zeros((len(patterns), 2))

        for ix, pattern in enumerate(patterns):
            consequent_match = np.equal(self.y , self.consequents[ix])
            pattern_firing_strength = antecedent_memberships[:, ix]

            if pattern_firing_strength[consequent_match].shape[0] > 0:
                    res[ix] = np.mean(pattern_firing_strength[consequent_match])
            

        return res


    def compute_aux_pattern_support(self) -> np.array:
        '''
        Computes the pattern support for each of the rules for each of the classes for the given X.
        Each pattern support firing strength is the result of the tnorm for all the antecedent memeberships,
        dvided by their number.

        :return: array of shape rules x 2
        '''
        if self.time_moments is None:
            antecedent_memberships = self.mrule_base.compute_firing_strenghts(
                self.X)
        else:
            antecedent_memberships = self.mrule_base.compute_firing_strenghts(
                self.X, self.time_moments)

        patterns = self._get_all_rules()
        n_classes = len(np.unique(self.y))

        if self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.t1:
            res = np.zeros((len(patterns), n_classes, ))
        elif self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.t2:
            res = np.zeros((len(patterns), n_classes, 2))
        elif self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.gt2:
            res = np.zeros((len(patterns), n_classes, 2))

        for con_ix in range(n_classes):
            for ix, pattern in enumerate(patterns):
                consequent_match = self.y == con_ix
                pattern_firing_strength = antecedent_memberships[:, ix]

                # / pattern_firing_strength.shape[0]
                if pattern_firing_strength[consequent_match].shape[0] > 0:
                    res[ix, con_ix] = np.mean(pattern_firing_strength[consequent_match])
                else:
                    res[ix, con_ix] = pattern_firing_strength[consequent_match]

        return res


    def _get_all_rules(self) -> list[rules.RuleSimple]:
        '''
        Returns a list of all the rules in the master rule base.

        :return: list of rules.
        '''
        res = []
        for jx in self.mrule_base.get_rules():
                res.append(jx)

        return res


    def compute_pattern_confidence(self) -> np.array:
        '''
        Computes the pattern confidence for each of the rules for the given X.
        Each pattern confidence is the normalized firing strength.

        :returns: array of shape 1 x rules 
        '''
        if self.time_moments is None:
            antecedent_memberships = self.mrule_base.compute_firing_strenghts(
            self.X, precomputed_truth=self.precomputed_truth)
        else:
            antecedent_memberships = self.mrule_base.compute_firing_strenghts(
            self.X, self.time_moments, precomputed_truth=self.precomputed_truth)

        patterns = self._get_all_rules()

        if self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.t1:
            res = np.zeros((len(patterns), ))
        elif self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.t2:
            res = np.zeros((len(patterns), 2))
        elif self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.gt2:
            res = np.zeros((len(patterns), 2))

        for ix, pattern in enumerate(patterns):
            antecedent_consequent_match = np.equal(self.y, self.consequents[ix])
            pattern_firing_strength = antecedent_memberships[:, ix]
            dem = np.sum(pattern_firing_strength)
            if dem == 0:
                res[ix] = 0
            else:
                res[ix] = np.sum(
                    pattern_firing_strength[antecedent_consequent_match]) / dem

        return res


    def compute_aux_pattern_confidence(self) -> np.array:
        '''
        Computes the pattern confidence for each of the rules for the given X.
        Each pattern confidence is the normalized firing strength.

        :returns: array of shape rules x classes
        '''
        if self.time_moments is None:
            antecedent_memberships = self.mrule_base.compute_firing_strenghts(
            self.X)
        else:
            antecedent_memberships = self.mrule_base.compute_firing_strenghts(
            self.X, self.time_moments)

        patterns = self._get_all_rules()
        n_classes = len(np.unique(self.y))
        if self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.t1:
            res = np.zeros((len(patterns), n_classes, ))
        elif self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.t2:
            res = np.zeros((len(patterns), n_classes, 2))
        elif self.mrule_base.fuzzy_type() == fs.FUZZY_SETS.gt2:
            res = np.zeros((len(patterns), n_classes, 2))

        for consequent in range(n_classes):
            for ix, pattern in enumerate(patterns):
                antecedent_consequent_match = self.y == consequent
                pattern_firing_strength = antecedent_memberships[:, ix]
                dem = np.sum(pattern_firing_strength)
                if dem == 0:
                    res[ix, consequent] = 0
                else:
                    res[ix, consequent] = np.sum(
                        pattern_firing_strength[antecedent_consequent_match]) / dem

        return res


    def dominance_scores(self) -> np.array:
        '''
        Returns the dominance score of each pattern for each rule.

        :return: array of shape rules x 2
        '''
        return self.compute_pattern_confidence() * self.compute_pattern_support()


    def association_degree(self) -> np.array:
        '''
        Returns the association degree of each rule for each sample.

        :return: vector of shape rules
        '''
        firing_strengths = self.mrule_base.compute_firing_strenghts(self.X)
        res = self.dominance_scores() * firing_strengths

        if (self.mrule_base[0].fuzzy_type() == fs.FUZZY_SETS.t2) or (self.mrule_base[0].fuzzy_type() == fs.FUZZY_SETS.gt2):
            res = np.mean(res, axis=2)

        return res


    def aux_dominance_scores(self) -> np.array:
        '''
        Returns the dominance score of each pattern for each rule.

        :return: array of shape rules x 2
        '''
        return self.compute_aux_pattern_confidence() * self.compute_aux_pattern_support()


    def add_rule_weights(self) -> None:
        '''
        Add dominance score field to each of the rules present in the master Rule Base.
        '''
        supports = self.compute_pattern_support()
        confidences = self.compute_pattern_confidence()
        scores = self.dominance_scores()

        aux_counter = 0
        rules = self.mrule_base.get_rules()
        for jx in range(len(rules)):
                rules[jx].score = scores[aux_counter]
                rules[jx].support = supports[aux_counter]
                rules[jx].confidence = confidences[aux_counter]
                
                aux_counter += 1
    

    def add_auxiliary_rule_weights(self) -> None:
        '''
        Add dominance score field to each of the rules present in the master Rule Base for each consequent.
        They are labeled as aux_score, aux_support and aux_confidence. (Because they are not the main rule weights)
        '''
        supports = self.compute_aux_pattern_support()
        confidences = self.compute_aux_pattern_confidence()
        scores = self.aux_dominance_scores()

        aux_counter = 0
        rules = self.mrule_base.get_rules()
        for jx in range(len(rules)):
                rules[jx].aux_score = scores[aux_counter]
                rules[jx].aux_support = supports[aux_counter]
                rules[jx].aux_confidence = confidences[aux_counter]
                
                aux_counter += 1


    def add_classification_metrics(self, X: np.array=None, y: np.array=None) -> None:
        '''
        Adds the accuracy of each rule in the master rule base. It also adds the f1, precision and recall scores.
        If X and y are None uses the train set.

        :param X: array of shape samples x features
        :param y: array of shape samples

        '''
        if X is not None:
            actual_X = X
            actual_y = y
        else:
            actual_X = self.X
            actual_y = self.y
        
        if not hasattr(self.mrule_base.get_rules()[0], 'score'):
            self.add_rule_weights()

        if self.time_moments is None:
            winning_rules = self.mrule_base._winning_rules(actual_X, precomputed_truth=self.precomputed_truth)
            preds = self.mrule_base.winning_rule_predict(actual_X, precomputed_truth=self.precomputed_truth)
        else:
            winning_rules = self.mrule_base._winning_rules(actual_X, self.time_moments)
            preds = self.mrule_base.winning_rule_predict(actual_X, self.time_moments)

        rules = self.mrule_base.get_rules()
        for jx in range(len(rules)):
                relevant_samples = winning_rules == jx
                if np.sum(relevant_samples) > 0:
                    relevant_labels = actual_y[relevant_samples]
                    relevant_preds = preds[relevant_samples]

                    rules[jx].accuracy = accuracy_score(relevant_labels, relevant_preds)
                else:
                    rules[jx].accuracy = 0.0
                
    
    def classification_eval(self) -> float:
        '''
        Returns the matthews correlation coefficient for a classification task using
        the rules evaluated.

        :return: mattews correlation coefficient. (float in [-1, 1])
        '''
        from sklearn.metrics import matthews_corrcoef
        self.add_rule_weights()
        preds = self.mrule_base.winning_rule_predict(self.X)

        return matthews_corrcoef(self.y, preds)


    def size_antecedents_eval(self, tolerance=0.1) -> float:
        '''
        Returns a score between 0 and 1, where 1 means that the rule base only contains almost no antecedents.

        0 means that the rule base contains all rules with more than {tolerance} DS, there are many of them and they have all possible antecedents.
        The more rules and antecedent per rules the lower this score is.

        :param tolerance: float in [0, 1]. The tolerance for the dominance score. Default 0.1
        :return: float in [0, 1] with the score.
        '''
        possible_rule_size = 0
        effective_rule_size = 0

        for rule_base in self.mrule_base.get_rulebases():
            if len(rule_base) > 0:
                for rule in rule_base:
                    rscore = np.mean(rule.score)
                    if rscore > tolerance:
                        possible_rule_size += len(rule.antecedents)
                        # No antecedents for this rule
                        if sum(np.array(rule.antecedents) != -1) == 0:
                            effective_rule_size += len(rule.antecedents)
                        else:
                            effective_rule_size += sum(
                                np.array(rule.antecedents) != -1)

            else:
                return 0.0  # If one consequent does not have rules, then we return 0.0

        try:
            rule_density = 1 - effective_rule_size / possible_rule_size  # Antecedents used
        except ZeroDivisionError:
            rule_density = 0.0

        return rule_density
    

    def effective_rulesize_eval(self, tolerance=0.1) -> float:
        '''
        Returns a score between 0 and 1, where 1 means that the rule base only contains almost no antecedents.

        0 means that the rule base contains all rules with more than {tolerance} DS, there are many of them and they have all possible antecedents.
        The more rules and antecedent per rules the lower this score is.

        :param tolerance: float in [0, 1]. The tolerance for the dominance score. Default 0.1
        :return: float in [0, 1] with the score.
        '''
        possible_rules = len(self.mrule_base.get_rules())
        effective_rules = 0

        for rule_base in self.mrule_base.get_rulebases():
            if len(rule_base) > 0:
                for rule in rule_base:
                    rscore = np.mean(rule.score)
                    if rscore > tolerance:
                        # No antecedents for this rule
                        if not np.all(np.equal(np.array(rule.antecedents), -1)):
                            effective_rules += 1
                        
            else:
                return 0.0  # If one consequent does not have rules, then we return 0.0

        try:
            rule_density =  effective_rules / possible_rules
        except ZeroDivisionError:
            rule_density = 0.0

        return rule_density


    def add_full_evaluation(self):
        '''
        Adds classification scores, both Dominance Scores and accuracy metrics, for each individual rule.
        '''
        self.add_rule_weights()
        self.add_classification_metrics()
        



# -*- coding: utf-8 -*-
"""
This is the source file that contains functions to compute centroids for the case of IV fuzzy sets,
which are commonly used when using the IV-T2 fuzzy sets.
"""
import numpy as np


def center_of_masses(z: np.array, w: np.array) -> np.array:
    '''
    Computes the ponderated centroid with normalized weighting.

    :param z: Vector of the referencial values.
    :param w: Vector of the fuzzy memberships.
    :return: The ponderated sum.
    '''
    return z @ w / np.sum(w)


def compute_centroid_t2_l(z: np.array, memberships: np.array) -> float:
    '''
    Computes the Karnik and Mendel algorithm to find the centroid of an IV fuzzy set.

    :param z: Vector of the referencial values.
    :param memberships: vector of the fuzzy memberships.
    :return: The centroid.
    '''
    centros = np.mean(memberships, axis=1)
    w = centros

    yhat = center_of_masses(z, w)

    yhat_2 = None

    while(yhat != yhat_2):
        try:
            k = np.argwhere((z - yhat) >= 0)[-1][0]
            k = min(len(centros)-1, k)
        except IndexError:
            k = 0

        # k_vector = np.argwhere((z - yhat) > 0)
        # k = k_vector[0][0] + 1
        w[0:k] = memberships[:, 1][:k]
        w[k:] = memberships[:, 0][k:]

        yhat_2 = yhat
        yhat = center_of_masses(z, w)

    return yhat


def compute_centroid_t2_r(z: np.array, memberships: np.array) -> float:
    '''
    Computes the Karnik and Mendel algorithm to find the right component of a centroid of an IV fuzzy set.

    :param z: Vector of the referencial values.
    :param memberships: vector of the fuzzy memberships.
    :return: The lowest membership of the centroid.
    '''
    centros = np.mean(memberships, axis=1)
    w = centros

    yhat = center_of_masses(z, w)

    yhat_2 = None

    while(yhat != yhat_2):
        try:
            k = np.argwhere((z - yhat) >= 0)[-1][0]
            k = min(len(centros)-1, k)
        except IndexError:
            k = 0

        w[0:k+1] = memberships[:, 0][:k+1]
        w[k+1:] = memberships[:, 1][k+1:]

        yhat_2 = yhat
        yhat = center_of_masses(z, w)

    return yhat


def consequent_centroid_r(antecedent_memberships: np.array, centroids_r: np.array) -> float:
    '''
    Computes the Karnik and Mendel algorithm to find the right component of a centroid of an IV fuzzy set
    using the centroids of the consequents and the antecedents memeberships.

    :param antecedent_memberships: M x 2 matrix. M rules and iv dimension (2). Vector of memberships.
    :param centroids_r: M sized vector. Contains the referencial values.
    :return: the IV centroid.
    '''

    memberships_left = antecedent_memberships[:, 0]
    memberships_right = antecedent_memberships[:, 1]
    initial_f = (memberships_left + memberships_right) / 2

    yhat = center_of_masses(centroids_r, initial_f)
    yhat2 = None

    while(yhat != yhat2):
        try:
            r_R = np.argwhere((yhat - centroids_r) >= 0)[-1][0]
            r_R = min(len(centroids_r)-1, r_R)
        except IndexError:
            r_R = 0

        new_memberships = initial_f
        new_memberships[0:r_R+1] = memberships_left[0:r_R+1]
        new_memberships[r_R+1:] = memberships_right[r_R+1:]

        yhat2 = yhat
        if np.sum(new_memberships) == 0:
            yhat = 0
        else:
            yhat = center_of_masses(centroids_r, new_memberships)

    return yhat2


def consequent_centroid_l(antecedent_memberships: np.array, centroids_l: np.array) -> float:
    '''
    Computes the Karnik and Mendel algorithm to find the left component of a centroid of an IV fuzzy set
    using the centroids of the consequents and the antecedent memberships.

    :param antecedent_memberships: M x 2 matrix. M rules and iv dimension (2). Vector of memberships.
    :param centroids_r: M sized vector. Contains the referencial values.
    :return: the IV centroid.
    '''

    memberships_left = antecedent_memberships[:, 0]
    memberships_right = antecedent_memberships[:, 1]
    initial_f = (memberships_left + memberships_right) / 2

    # (memberships_right * centroids_r) / np.sum(memberships_right)
    yhat = center_of_masses(centroids_l, initial_f)
    yhat2 = -100

    #R = np.where(yhat - centroids_r)[0]
    while(yhat != yhat2):
        try:
            r_R = np.argwhere((yhat - centroids_l) >= 0)[-1][0]
            r_R = min(len(centroids_l)-1, r_R)
        except IndexError:
            r_R = 0

        new_memberships = initial_f
        new_memberships[0:r_R+1] = memberships_right[0:r_R+1]
        new_memberships[r_R+1:] = memberships_left[r_R+1:]
        yhat2 = yhat
        if np.sum(new_memberships) == 0:
            yhat = 0
        else:
            yhat = center_of_masses(centroids_l, new_memberships)

    return yhat2


def compute_centroid_iv(z: np.array, memberships: np.array) -> np.array:
    '''
    Computes Karnik and Mendel algorithm to find the centroid of a iv fuzzy set. 

    :param z: Vector of the referencial values.
    :param memberships: vector of the fuzzy memberships.
    :return: The centroid.
    '''
    res = np.zeros((2,))
    res[0] = compute_centroid_t2_l(z, memberships)
    res[1] = compute_centroid_t2_r(z, memberships)
    return res


def consequent_centroid(antecedent_memberships: np.array, centroids: np.array) -> np.array:
    '''
    Computes Karnik and Mendel algorithm to find the centroid of a consequent iv fuzzy set given the centroids of the consequents
    and the antecedents memberships. 

    :param antecedent_memberships: M x 2 matrix. M rules and iv dimension (2). Vector of memberships.
    :param centroids: M x 2 matrix. M rules and iv dimension (2). Vector of centroids.
    :return: The centroid.
    '''
    centroids_l = centroids[:, 0]
    centroids_r = centroids[:, 1]
    res = np.zeros((2,))

    res[0] = consequent_centroid_l(antecedent_memberships, centroids_l)
    res[1] = consequent_centroid_r(antecedent_memberships, centroids_r)

    return res






import numpy as np
import ex_fuzzy
import sys
sys.path.append('../ex_fuzzy/')

def test_eval_fuzzy_model():
    '''
    Test the eval_fuzzy_model function.

    If it works, it should print the following:
    Accuracy: 0.5 aprox
    Matthews correlation coefficient: 0.0 aprox
    '''
    sample_size = 1000
    sample = np.random.random_sample((sample_size, 5))
    targets = np.random.randint(0, 2, sample_size)

    model = ex_fuzzy.evolutionary_fit.BaseFuzzyRulesClassifier(10, 3, ex_fuzzy.fuzzy_sets.FUZZY_SETS.t1, verbose=False, tolerance=0.0, n_linguist_variables=3)
    model.fit(sample, targets)

    ex_fuzzy.eval_tools.eval_fuzzy_model(model, sample, targets, sample, targets, 
                        plot_rules=False, print_rules=False, plot_partitions=False)




# -*- coding: utf-8 -*-
"""
Functions that are not fuzzy-specific, but util for some computations. 
Dedicated mostly to compute quantiles for fuzzy partitions.

"""
import numpy as np
import pandas as pd

try:
    from . import fuzzy_sets as fs
    from . import maintenance as mnt
    from . import temporal
    from . import rules
    from . import eval_rules as evr
except ImportError:
    import fuzzy_sets as fs
    import maintenance as mnt
    import temporal
    import rules
    import eval_rules as evr


from sklearn.model_selection import train_test_split


def quartile_compute(x: np.array) -> list[float]:
    '''
    Computes the quartiles for each feature.

    :param x: array samples x features
    :return: list of quartiles for each feature
    '''
    return np.quantile(x, [0, 0.25, 0.5, 1], axis=0)


def fixed_quantile_compute(x: np.array) -> list[float]:
    '''
    Computes a series of quantiles for each feature in numpy array.
    Quantiles: [0, 0.20, 0.30, 0.45, 0.55, 0.7, 0.8, 1]

    :param x: array samples x features
    :return: list of quantiles for each feature
    '''
    return np.quantile(x, [0, 0.20, 0.30, 0.45, 0.55, 0.7, 0.8, 1], axis=0)


def partition3_quantile_compute(x: np.array) -> list[float]:
    '''
    Computes a series of quantiles partitioning the variable in 3 cases.

    Quantiles: [0.00, 0.20, 0.50, 0.80, 1.00]

    :param x: array samples x features
    :return: list of quantiles for each feature
    '''
    return np.quantile(x, [0, 0.20, 0.50, 0.80, 1.00], axis=0)


def t1_simple_partition(x: np.array) -> np.array:
    '''
    Partitions the fuzzy variable in four trapezoidal memberships.

    :param x: numpy array, vector of shape (samples, ).
    :return: numpy array, vector of shape (variables, 4, 4).
    '''
    n_partitions = 4
    trap_memberships_size = 4
    quantile_numbers = fixed_quantile_compute(x)

    partition_parameters = np.zeros(
        (x.shape[1], n_partitions, trap_memberships_size))
    for partition in range(n_partitions):
        if partition == 0:
            partition_parameters[:, partition, 0] = quantile_numbers[0]
            partition_parameters[:, partition, 1] = quantile_numbers[0]
            partition_parameters[:, partition, 2] = quantile_numbers[1]
            partition_parameters[:, partition, 3] = quantile_numbers[2]
        elif partition == n_partitions - 1:
            partition_parameters[:, partition, 0] = quantile_numbers[-3]
            partition_parameters[:, partition, 1] = quantile_numbers[-2]
            partition_parameters[:, partition, 2] = quantile_numbers[-1]
            partition_parameters[:, partition, 3] = quantile_numbers[-1]
        else:
            pointer = 1 if partition == 1 else 4
            partition_parameters[:, partition, 0] = quantile_numbers[pointer]
            partition_parameters[:, partition,
                                 1] = quantile_numbers[pointer + 1]
            partition_parameters[:, partition,
                                 2] = quantile_numbers[pointer + 2]
            partition_parameters[:, partition,
                                 3] = quantile_numbers[pointer + 3]

    return partition_parameters


def t1_three_partition(x: np.array) -> np.array:
    '''
    Partitions the fuzzy variable in three trapezoidal memberships.

    :param x: numpy array, vector of shape (samples, ).
    :return: numpy array, vector of shape (variables, 3, 4).
    '''
    n_partitions = 3
    trap_memberships_size = 4
    quantile_numbers = partition3_quantile_compute(x)

    partition_parameters = np.zeros(
        (x.shape[1], n_partitions, trap_memberships_size))
    for partition in range(n_partitions):
        if partition == 0:
            partition_parameters[:, partition, 0] = quantile_numbers[0]
            partition_parameters[:, partition, 1] = quantile_numbers[0]
            partition_parameters[:, partition, 2] = quantile_numbers[1]
            partition_parameters[:, partition, 3] = quantile_numbers[2]
        elif partition == 1:
            partition_parameters[:, partition, 0] = quantile_numbers[1]
            partition_parameters[:, partition, 1] = (
                quantile_numbers[1] + quantile_numbers[2]) / 2
            partition_parameters[:, partition, 2] = (
                quantile_numbers[3] + quantile_numbers[2]) / 2
            partition_parameters[:, partition, 3] = quantile_numbers[3]
        else:
            partition_parameters[:, partition, 0] = quantile_numbers[2]
            partition_parameters[:, partition, 1] = quantile_numbers[3]
            partition_parameters[:, partition, 2] = quantile_numbers[4]
            partition_parameters[:, partition, 3] = quantile_numbers[4]

    return partition_parameters


def t2_simple_partition(x: np.array) -> np.array:
    '''
    Partitions the fuzzy variable in three trapezoidal memberships.

    :param x: numpy array, vector of shape (samples, ).
    :return: numpy array, vector of shape (variables, 3, 4, 2).
    '''
    n_partitions = 3
    trap_memberships_size = 4
    quantile_numbers = partition3_quantile_compute(x)

    partition_parameters = np.zeros(
        (x.shape[1], n_partitions, trap_memberships_size, 2))
    for partition in range(n_partitions):
        if partition == 0:
            partition_parameters[:, partition, 0, 1] = quantile_numbers[0]
            partition_parameters[:, partition, 1, 1] = quantile_numbers[0]
            partition_parameters[:, partition, 2, 1] = quantile_numbers[1]
            partition_parameters[:, partition, 3, 1] = quantile_numbers[2]

            partition_parameters[:, partition, 0, 0] = quantile_numbers[0]
            partition_parameters[:, partition, 1, 0] = quantile_numbers[0]
            partition_parameters[:, partition, 2, 0] = quantile_numbers[1]
            partition_parameters[:, partition, 3, 0] = quantile_numbers[1] + \
                0.9 * (quantile_numbers[2] - quantile_numbers[1])

        elif partition == 1:
            partition_parameters[:, partition, 0, 1] = quantile_numbers[1]
            partition_parameters[:, partition, 1, 1] = (
                quantile_numbers[1] + quantile_numbers[2]) / 2
            partition_parameters[:, partition, 2, 1] = (
                quantile_numbers[2] + quantile_numbers[3]) / 2
            partition_parameters[:, partition, 3, 1] = quantile_numbers[3]

            partition_parameters[:, partition, 0, 0] = quantile_numbers[1] + \
                0.1 * (quantile_numbers[2] - quantile_numbers[1])
            partition_parameters[:, partition, 1, 0] = (
                quantile_numbers[1] + quantile_numbers[2]) / 2
            partition_parameters[:, partition, 2, 0] = (
                quantile_numbers[3] + quantile_numbers[2]) / 2
            partition_parameters[:, partition, 3, 0] = quantile_numbers[2] + \
                0.9 * (quantile_numbers[3] - quantile_numbers[2])

        else:
            partition_parameters[:, partition, 0, 1] = quantile_numbers[2]
            partition_parameters[:, partition, 1, 1] = quantile_numbers[3]
            partition_parameters[:, partition, 2, 1] = quantile_numbers[4]
            partition_parameters[:, partition, 3, 1] = quantile_numbers[4]

            partition_parameters[:, partition, 0, 0] = quantile_numbers[2] + \
                0.1 * (quantile_numbers[3] - quantile_numbers[2])
            partition_parameters[:, partition, 1, 0] = quantile_numbers[3]
            partition_parameters[:, partition, 2, 0] = quantile_numbers[4]
            partition_parameters[:, partition, 3, 0] = quantile_numbers[4]

    return partition_parameters


def t1_fuzzy_partitions_dataset(x0: np.array) -> list[fs.fuzzyVariable]:
    '''
    Partitions the dataset features into different fuzzy variables. Parameters are prefixed.
    Use it for simple testing and initial solution.

    :param x: numpy array|pandas dataframe, shape samples x features.
    :return: list of fuzzy variables.
    '''
    tripartition_names = ['Low', 'Medium', 'High']
    try:
        fv_names = x0.columns
        x = x0.values
    except AttributeError:
        fv_names = [str(ix) for ix in range(x0.shape[1])]
        x = x0

    mins = np.min(x, axis=0)
    maxs = np.max(x, axis=0)
    fz_memberships = t1_three_partition(x)
    res = []
    for fz_parameter in range(fz_memberships.shape[0]):
        fzs = [fs.FS(tripartition_names[ix], fz_memberships[fz_parameter, ix, :], [
                     mins[fz_parameter], maxs[fz_parameter]]) for ix in range(fz_memberships.shape[1])]
        res.append(fs.fuzzyVariable(fv_names[fz_parameter], fzs))

    return res


def t2_fuzzy_partitions_dataset(x0: np.array) -> list[fs.fuzzyVariable]:
    '''
    Partitions the dataset features into different fuzzy variables using iv fuzzy sets. Parameters are prefixed.
    Use it for simple testing and initial solution.

    :param x: numpy array|pandas dataframe, shape samples x features.
    :return: list of fuzzy variables.
    '''
    tripartition_names = ['Low', 'Medium', 'High']
    try:
        fv_names = x0.columns
        x = x0.values
    except AttributeError:
        fv_names = [str(ix) for ix in range(x0.shape[1])]
        x = x0

    mins = np.min(x, axis=0)
    maxs = np.max(x, axis=0)
    fz_memberships = t2_simple_partition(x)
    res = []
    for fz_parameter in range(fz_memberships.shape[0]):
        fzs = [fs.IVFS(tripartition_names[ix], fz_memberships[fz_parameter, ix, :, 0], fz_memberships[fz_parameter, ix, :, 1], [
                       mins[fz_parameter], maxs[fz_parameter]], lower_height=0.8) for ix in range(fz_memberships.shape[1])]
        res.append(fs.fuzzyVariable(fv_names[fz_parameter], fzs))

    return res


def gt2_fuzzy_partitions_dataset(x0: np.array, resolution_exp:int=1) -> list[fs.fuzzyVariable]:
    '''
    Partitions the dataset features into different fuzzy variables using gt2 fuzzy sets. Parameters are prefixed.
    Use it for simple testing and initial solution.

    :param x: numpy array|pandas dataframe, shape samples x features.
    :param resolution_exp: exponent of the resolution of the partition. Default is -2, which means 0.01. (Number of significant decimals)
    :return: list of fuzzy variables.
    '''
    try:
        fv_names = x0.columns
        x = x0.values
    except AttributeError:
        fv_names = [str(ix) for ix in range(x0.shape[1])]
        x = x0

    mins = np.min(x, axis=0)
    maxs = np.max(x, axis=0)
    iv_simple_partition = t2_fuzzy_partitions_dataset(x)
    resolution = 10.0**-np.abs(resolution_exp)
    res = []
    # We iterate through all possible variables
    for ix_var, fz_var in enumerate(iv_simple_partition):
        domain_resolution = np.arange(
            mins[ix_var], maxs[ix_var] + resolution, resolution)
        fzs = []
        for ix_lv, fz_lv in enumerate(fz_var.get_linguistic_variables()):
            memberships = fz_lv.membership(domain_resolution)
            fs_domain = {}
            for ix_z, x in enumerate(domain_resolution):
                membership_z = memberships[ix_z]
                fs_domain[x] = fs.FS(str(x), [membership_z[0], np.mean(
                   membership_z), np.mean(membership_z), membership_z[1]], [0.0, 1.0])

            fzs.append(fs.GT2(fz_lv.name, fs_domain, [
                       mins[ix_var], maxs[ix_var]], significant_decimals=np.abs(resolution_exp), unit_resolution=0.01))

        res.append(fs.fuzzyVariable(fv_names[ix_var], fzs))

    return res


def construct_partitions(X : np.array, fz_type_studied:fs.FUZZY_SETS, categorical_mask: np.array=None) -> list[fs.fuzzyVariable]:
    '''
    Returns a list of linguistic variables according to the kind of fuzzy specified.

    :param X: numpy array|pandas dataframe, shape samples x features.
    :param fz_type_studied: fuzzy set type studied.
    :param categorial_mask: a boolean mask vector that indicates for each variables if its categorical or not.
    '''
    if mnt.save_usage_flag:
        mnt.usage_data[mnt.usage_categories.Funcs]['precompute_labels'] += 1
        mnt.usage_data[mnt.usage_categories.FuzzySets][fz_type_studied.name] += 1

    if isinstance(X, pd.DataFrame):
        feat_names = X.columns
        X = X.values
    else:
        feat_names = [str(ix) for ix in range(X.shape[1])]

    # Get the X dataframe without the categorical variables
    if categorical_mask is not None:
        X_numerical = X[:, np.logical_not(categorical_mask)]
    else:
        X_numerical = X

    if fz_type_studied == fs.FUZZY_SETS.t1:
        precomputed_partitions = t1_fuzzy_partitions_dataset(X_numerical)
    elif fz_type_studied == fs.FUZZY_SETS.t2:
        precomputed_partitions = t2_fuzzy_partitions_dataset(X_numerical)
    elif fz_type_studied == fs.FUZZY_SETS.gt2:
        precomputed_partitions = gt2_fuzzy_partitions_dataset(X_numerical)


    if categorical_mask is not None:
        categorical_partition = {}
        for ix, elem in enumerate(categorical_mask):
            if elem:
                if isinstance(X, pd.DataFrame):
                    name = X.columns[ix]
                else:
                    name = str(ix)
                cat_var = construct_crisp_categorical_partition(np.array(X)[:, ix], name, fz_type_studied)

                categorical_partition[name] = cat_var

        # Reorder the partitions so that they follow the same order as in the original X
        precomputed_partitions_aux = []
        for ix, elem in enumerate(categorical_mask):
            if isinstance(X, pd.DataFrame):
                name = X.columns[ix]
            else:
                name = str(ix)

            if elem:
                precomputed_partitions_aux.append(categorical_partition[name])
            else:
                precomputed_partitions_aux.append(precomputed_partitions.pop(0))

        precomputed_partitions = precomputed_partitions_aux

    
    for ix, partition in enumerate(precomputed_partitions):
        partition.name = feat_names[ix]

    return precomputed_partitions


def construct_crisp_categorical_partition(x: np.array, name: str, fz_type_studied: fs.FUZZY_SETS) -> fs.fuzzyVariable:
    '''
    Creates a fuzzy variable for a categorical feature. 

    :param x: array with values of the categorical variable.
    :param name of the fuzzy variable.
    :param fz_type_studied: fuzzy set type studied.
    :return: a fuzzy variable that works as a categorical crips variable (each fuzzy set is 1 exactly on each class value, and 0 on the rest)
    '''
    possible_values = np.unique(x)
    fuzzy_sets = []

    # Create a fuzzy sets for each possible value
    for ix, value in enumerate(possible_values):

        if fz_type_studied == fs.FUZZY_SETS.t1:
            aux = fs.categoricalFS(str(value), value)
        elif fz_type_studied == fs.FUZZY_SETS.t2 or fz_type_studied == fs.FUZZY_SETS.gt2:
            aux = fs.categoricalIVFS(str(value), np.unique(x))

        fuzzy_sets.append(aux)

    return fs.fuzzyVariable(name, fuzzy_sets)
    

def construct_conditional_frequencies(X: np.array, discrete_time_labels: list[int], initial_ffss: list[fs.FS]):
    '''
    Computes the conditional temporal function for a set of fuzzy sets according to their variation in time.

    :param X: numpy array, shape samples x features.
    :param discrete_time_labels: discrete time labels.
    :param initial_fs: initial fuzzy set list.
    :return: conditional frequencies. Array shape (time steps, initial fuzzy sets)
    '''
    obs = X.shape[0]
    discrete_time_labels = np.array(discrete_time_labels)
    memberships = np.zeros((obs, len(initial_ffss)))
    for ix, fset in enumerate(initial_ffss):
        if fset.type() == fs.FUZZY_SETS.t2:
            memberships[:, ix] = np.mean(fset.membership(X), axis=1)
        elif fset.type() == fs.FUZZY_SETS.gt2:
            memberships[:, ix] = np.mean(np.squeeze(fset._alpha_reduction(fset.membership(X))), axis=1)
        else:
            memberships[:, ix] = fset.membership(X)
    
    max_memberships = np.argmax(memberships, axis=1)
    res = np.zeros((len(np.unique(discrete_time_labels)), len(initial_ffss)))

    for time in range(len(np.unique(discrete_time_labels))):

        relevant_memberships = max_memberships[time == discrete_time_labels]
        fs_winner_counter = np.zeros(len(initial_ffss))
        for ix, fset in enumerate(initial_ffss):
            fs_winner_counter[ix] = np.sum(relevant_memberships == ix)
        
        res[time, :] = fs_winner_counter
    
    return res / (np.max(res, axis=0) + 1e-6)


def classify_temp(dates: pd.DataFrame, cutpoints: tuple[str, str], time: str) -> np.array:
    '''
    Classifies a set of dates according to the temporal cutpoints. Uses {time} as a the time resolution.
    Returns an array where true values are those values contained between those two date points.

    :param dates: data observations to cut. 
    :param cutpoints: points to check.
    :param time: time field to use as the criteria.
    :return: boolean array. True values are those contained between the cutpoints.
    '''

    def extract_hour(row):
        return row.__getattribute__(time)
    
    hours = pd.to_datetime(dates['date']).apply(extract_hour)

    cutpoint_series_0 =  pd.to_datetime(pd.Series([cutpoints[0]] * len(dates)))
    cutpoint_series_0.index = dates.index
    hours0 = cutpoint_series_0.apply(extract_hour)

    cutpoint_series_1 =  pd.to_datetime(pd.Series([cutpoints[1]] * len(dates)))
    cutpoint_series_1.index = dates.index
    hours1 = cutpoint_series_1.apply(extract_hour)

    condicion1 = hours >= hours0
    condicion2 = hours <= hours1

    return np.array(np.logical_and(condicion1, condicion2))


def assign_time(a: np.array, observations: list[np.array]) -> int:
    '''
    Assigns a temporal moment to a set of observations.
    
    :param a: array of boolean values.
    :param observations: list of boolean arrays with the corresponding timestamps.
    :return: the index of the correspondent time moment for the a-th observation.
    :raises: ValueError if a is not timestamped in any of the observation arrays.'''
    for ix, obs in enumerate(observations):
        if obs[a]:
            return ix
    
    raise ValueError('No temporal moment assigned')
        

def create_tempVariables(X_train: np.array, time_moments: np.array, precomputed_partitions: list[fs.fuzzyVariable]) -> list[temporal.temporalFS]:
    '''
    Creates a list of temporal fuzzy variables.

    :param X_train: numpy array, shape samples x features.
    :param time_moments: time moments. Array shape (samples,). Each value is an integer denoting the n-th time moment of that observation.
    :param precomputed_partitions: precomputed partitions for each feature.
    :return: list of temporal fuzzy variables.
    '''
    temp_partitions = []
    for ix in range(X_train.shape[1]):
        feat_conditional = construct_conditional_frequencies(X_train[:, ix], time_moments, initial_ffss=precomputed_partitions[ix])
        temp_fs_list = []
        for vl in range(feat_conditional.shape[1]):
            vl_temp_fs = temporal.temporalFS(precomputed_partitions[ix][vl], feat_conditional[:, vl])
            temp_fs_list.append(vl_temp_fs)

        temp_fs_variable = temporal.temporalFuzzyVariable(precomputed_partitions[ix].name, temp_fs_list)
        temp_partitions.append(temp_fs_variable)

    return temp_partitions


def create_multi_tempVariables(X_train: np.array, time_moments: np.array, fuzzy_type: fs.FUZZY_SETS) -> list[list[temporal.temporalFS]]:
    '''
    Creates a of list of lists of temporal fuzzy variables. Each corresponds to a fuzzy partition in a different moment in time.
    (So, instead of having one vl for all time moments, you have one different for each time moment that represents the same idea)

    :param X_train: numpy array, shape samples x features.
    :param time_moments: time moments. Array shape (samples,). Each value is an integer denoting the n-th time moment of that observation.
    :param precomputed_partitions: precomputed partitions for each feature.
    :return: list of lists of temporal fuzzy variables.
    '''
    temp_partitions = []

    unique_time_moments = np.unique(time_moments)
    for time in unique_time_moments:
        X_obs = X_train[time_moments == time, :]
        precomputed_partitions = construct_partitions(X_obs, fuzzy_type)

        temp_partitions.append(create_tempVariables(X_obs, time_moments[time_moments == time], precomputed_partitions))
    
    return temp_partitions


def temporal_cuts(X: pd.DataFrame, cutpoints: list[tuple[str, str]], time_resolution: str='hour') -> list[np.array]:
    '''
    Returns a list of boolean indexes for each temporal moment. Performs the cuts between time steps using the cutpoints list.

    :param X: data observations to cut in temrporal moments.
    :param temporal_moments: list of temporal moments to cut.
    :param cutpoints: list of tuples with the cutpoints for each temporal moment.
    :param time_resolution: time field to use as the criteria.
    :return: list of boolean arrays. True values are those contained between the cutpoints in each moment.
    '''

    res = []
    for ix, cutpoint in enumerate(cutpoints):
        observations = classify_temp(X, cutpoint, time=time_resolution)
        res.append(observations)
    
    return res


def temporal_assemble(X: np.array, y:np.array, temporal_moments: list[np.array]):
    '''
    Assembles the data in the temporal moments in order to have partitions with balanced time moments in each one.
    
    :param X: data observations.
    :param y: labels.
    :param temporal_moments: list of boolean arrays. True values are those contained between the cutpoints in each moment.
    :return: tuple of lists of data and labels for each temporal moment.
        First tuple is: X_train, X_test, y_train, y_test
        Second tuple is: train temporal moments, test temporal moments.
    '''
    moments_partitions = []
    train_temporal_boolean_markers = []
    test_temporal_boolean_markers = []
    train_counter = 0
    test_counter = 0

    for ix, temporal_moment in enumerate(temporal_moments):
        X_train, X_test, y_train, y_test = train_test_split(X[temporal_moment], y[temporal_moment], test_size=0.33, random_state=0)
        moments_partitions.append((X_train, X_test, y_train, y_test))
    
    if isinstance(X_train,(pd.core.series.Series,pd.DataFrame)):
        X_train = pd.concat([moments_partitions[ix][0] for ix in range(len(moments_partitions))])
        X_test =  pd.concat([moments_partitions[ix][1] for ix in range(len(moments_partitions))])
        y_train = np.concatenate([moments_partitions[ix][2] for ix in range(len(moments_partitions))])
        y_test = np.concatenate([moments_partitions[ix][3] for ix in range(len(moments_partitions))])
    else:
        X_train = np.concatenate([moments_partitions[ix][0] for ix in range(len(moments_partitions))])
        X_test =  np.concatenate([moments_partitions[ix][1] for ix in range(len(moments_partitions))])
        y_train = np.concatenate([moments_partitions[ix][2] for ix in range(len(moments_partitions))])
        y_test = np.concatenate([moments_partitions[ix][3] for ix in range(len(moments_partitions))])

    for ix, temporal_moment in enumerate(temporal_moments):
        # Believe, this makes sense to avoid rounding errrors in the size of the final vector
        _, _, y_train0, y_test0 = train_test_split(X[temporal_moment], y[temporal_moment], test_size=0.33, random_state=0)

        train_moment_observations = np.zeros((X_train.shape[0]))
        train_moment_observations[train_counter:train_counter+len(y_train0)] = 1
        train_counter += len(y_train0)
        train_temporal_boolean_markers.append(train_moment_observations)

        test_moment_observations = np.zeros((X_test.shape[0]))
        test_moment_observations[test_counter:test_counter+len(y_test0)] = 1
        test_counter += len(y_test0)
        test_temporal_boolean_markers.append(test_moment_observations)

    
    return [X_train, X_test, y_train, y_test], [train_temporal_boolean_markers, test_temporal_boolean_markers]


def extend_fuzzy_sets_enum(new_fuzzy_sets_enum: fs.FUZZY_SETS) -> list[fs.FUZZY_SETS]:
    '''
    Extends the fuzzy sets enum with additional types.

    :param fuzzy_sets_enum: fuzzy sets enum.
    :return: extended fuzzy sets enum.
    '''
    import enum
    NEW_FUZZY_SETS = enum.Enum(
        "FUZZY_SETS",
        [(es.name, es.value) for es in fs.FUZZY_SETS] + [(es.name, es.value) for es in new_fuzzy_sets_enum]
        )
    fs.FUZZY_SETS = NEW_FUZZY_SETS


def mcc_loss(ruleBase: rules.RuleBase, X:np.array, y:np.array, tolerance:float, alpha:float=0.99, beta:float=0.0125, gamma:float=0.0125, precomputed_truth=None) -> float:

        '''
        Fitness function for the optimization problem. Uses only the MCC, ignores the size penalization terms.


        :param ruleBase: RuleBase object
        :param X: array of train samples. X shape = (n_samples, n_features)
        :param y: array of train labels. y shape = (n_samples,)
        :param tolerance: float. Tolerance for the size evaluation.
        :param alpha: ignored.
        :param beta: ignored.
        :param gamma: ignored.
        :return: float. Fitness value.
        '''
        ev_object = evr.evalRuleBase(ruleBase, X, y, precomputed_truth=precomputed_truth)
        ev_object.add_rule_weights()

        score_acc = ev_object.classification_eval()
        
        return score_acc


'''
Expansion of the base fuzzy sets, adding temporal fuzzy sets.

Contains functions to model the temporal fuzzy sets, temporal fuzzy variables and temporal rule bases.
It also contains functions to evaluate the fuzzy rulebases obtained.

'''
import enum

import numpy as np
import pandas as pd

from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score, confusion_matrix, matthews_corrcoef
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.core.problem import Problem
from pymoo.optimize import minimize
from pymoo.operators.sampling.rnd import IntegerRandomSampling
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PolynomialMutation
from pymoo.core.variable import Integer
from multiprocessing.pool import ThreadPool
from pymoo.core.problem import StarmapParallelization

try:
    from . import fuzzy_sets as fs
    from . import maintenance as mnt
    from . import rules as rl
    from . import evolutionary_fit as evf
    from . import vis_rules
    from . import eval_rules as evr
except:
    import fuzzy_sets as fs
    import maintenance as mnt
    import rules as rl
    import evolutionary_fit as evf
    import vis_rules
    import eval_rules as evr


TMP_FUZZY_SETS = enum.Enum(
    "NEW_FUZZY_SETS",
    ['temporal', 'temporal_t2', 'temporal_gt2']
)
NEW_FUZZY_SETS = enum.Enum(
    "FUZZY_SETS",
    [(es.name, es.value) for es in fs.FUZZY_SETS] + [(es.name, es.name) for es in TMP_FUZZY_SETS]
)

fs.FUZZY_SETS = NEW_FUZZY_SETS
### DEFINE THE FUZZY SET ####
class temporalFS(fs.FS):
    '''
    Class to implement temporal fuzzy sets.
    '''

    def __init__(self, std_fuzzy_set: fs.FS, conditional_variable: np.array) -> None:
        '''
        Creates a temporal fuzzy set.

        :param std_fuzzy_set: FS. Standard fuzzy set that contains the non-temporal aware memberhsip function.
        :param conditional_variable: np.array. The variable that expresses the different temporal moments. Shape (time discrete moments, ).
        '''
        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.FuzzySets][self.type().name] += 1
            
        self.std_set = std_fuzzy_set
        self.tmp_function = conditional_variable
        self.time = None
        self.name = std_fuzzy_set.name

        if std_fuzzy_set.type() == fs.FUZZY_SETS.t1:
            self.membership_parameters = std_fuzzy_set.membership_parameters
        elif std_fuzzy_set.type() == fs.FUZZY_SETS.t2:
            self.secondMF_upper = std_fuzzy_set.secondMF_upper
            self.secondMF_lower = std_fuzzy_set.secondMF_lower
        elif std_fuzzy_set.type() == fs.FUZZY_SETS.gt2:
            self.secondary_memberships = std_fuzzy_set.secondary_memberships
            self.alpha_cuts = std_fuzzy_set.alpha_cuts


    def membership(self, input: np.array, time: int=None) -> np.array:
        '''
        Computes the membership of each sample and in each time for the fs.

        :param input: array temporal_time x samples.
        :time: int. Time moment to compute the membership. If none, looks for a fixed time
        :return: array temporal_time x samples.
        '''
        if time is None:
            assert self.time is not None, 'Temporal fuzzy set has no fixed time. Please, fix a time or provide a time to compute the membership.'
            time = self.time

        return self.std_set.membership(input) * self.tmp_function[time]


    def type(self) -> fs.FUZZY_SETS:
        '''
        Returns the type of the fuzzy set. (temporal)
        '''
        return fs.FUZZY_SETS.temporal
    

    def inside_type(self) -> fs.FUZZY_SETS:
        '''
        Returns the type of the og fuzzy set computed before the time dependency.
        '''
        return self.std_set.type()
    

    def fix_time(self, time: int) -> None:
        '''
        Fixes the time of the temporal fuzzy set.

        :param time: int. Time moment to fix.
        :return: FS. Fuzzy set with the fixed time.
        '''
        self.time = time



#### DEFINE THE FUZZY VARIABLE ####
class temporalFuzzyVariable(fs.fuzzyVariable):
    '''
    Class to implement a temporal fuzzy variable.
    '''

    def __init__(self, name: str, fuzzy_sets: list[temporalFS]) -> None:
        '''
        Creates a temporal fuzzy variable.

        :param str: name of the variable.
        :param fuzzy_sets: list of the fuzzy sets pre-time dependencies.
        '''
        self.linguistic_variables = []
        self.name = name
        self.time = None
        for ix, fs in enumerate(fuzzy_sets):
            self.linguistic_variables.append(fs)

        self.fs_type = self.linguistic_variables[0].type()


    def fix_time(self, time: int) -> None:
        '''
        Fixes the time of the temporal fuzzy variable.

        :param time: int. Time moment to fix.
        '''
        self.time = time


    def compute_memberships(self, x: np.array, time: int=None) -> dict:
        '''
        Computes the membership to each of the FS in the fuzzy variables.

        :param x: numeric value or array. Computes the membership to each of the IVFS in the fuzzy variables.
        :param time: int. Time moment to compute the membership.
        :return: list of floats. Membership to each of the FS in the fuzzy variables.
        '''
        if time is None:
            assert self.time is not None, 'Temporal fuzzy variable has no fixed time. Please, fix a time or provide a time to compute the membership.'
            time = self.time

        res = []

        for fuzzy_set in self.linguistic_variables:
            res.append(fuzzy_set.membership(x, time))

        return res
    

    def n_time_moments(self) -> int:
        '''
        Returns the number of time moments of the temporal fuzzy variable.

        :return: int. Number of time moments of the temporal fuzzy variable.
        '''
        return self.linguistic_variables[0].tmp_function.shape[0]


#### DEFINE THE RULE BASE WITH TEMPORAL DEPENDENCIES ####
class temporalMasterRuleBase(rl.MasterRuleBase):
    '''
    This class is a temporal extension of the MasterRuleBase class. It includes a list of
    rule bases for each time step.
    '''
    def __init__(self, rule_base: list[rl.MasterRuleBase], time_step_names: list[str]=None):
        '''
        Constructor of the temporalMasterRuleBase class.

        :param rule_base: list of rule bases.
        :param time_steps: number of time steps.
        '''
        super().__init__(rule_base)
        self.time_steps = np.arange(len(rule_base))
        self.time_mrule_bases = rule_base

        if time_step_names is None:
            self.time_step_names = [str(x) for x in self.time_steps]
        else:
            self.time_step_names = time_step_names
        
        for ix, mrb in enumerate(rule_base):
            for jx, rb in enumerate(mrb):
                for antecedent in rb.antecedents:
                    antecedent.fix_time(ix)
        
        self.antecedents = rule_base[0].antecedents
                     

    def add_rule(self, rule: rl.RuleSimple, consequent: int, time_step: int) -> None:
        '''
        Adds a rule to the rule base of the given consequent.

        :param rule: rule to add.
        :param consequent: index of the rule base to add the rule.
        :param time_step: time step of the rule base to add the rule.
        '''
        self.time_mrule_bases[time_step][consequent].add_rule(rule)


    def get_rulebase_matrix(self) -> list[np.array]:
        '''
        Returns a list with the rulebases for each antecedent in matrix format.

        :return: list with the rulebases for each antecedent in matrix format.
        '''
        return [a.get_rulebase_matrix() for x in self.time_mrule_bases for a in x]


    def get_scores(self) -> np.array:
        '''
        Returns the dominance score for each rule in all the rulebases.

        :return: array with the dominance score for each rule in all the rulebases.
        '''
        res = []
        for rule_bases in self.time_mrule_bases:    
            for rb in rule_bases:
                res.append(rb.scores())

        res = [x for x in res if len(x) > 0]

        return np.concatenate(res, axis=0)


    def compute_firing_strenghts(self, X: np.array, time_moments: list[int]) -> np.array:
        '''
        Computes the firing strength of each rule for each sample.

        :param X: array with the values of the inputs.
        :return: array with the firing strength of each rule for each sample.
        '''
        aux = []
        time_moments = np.array(time_moments)

        for time_moment, rule_bases in enumerate(self.time_mrule_bases):
            actual_moment = np.equal(time_moments, time_moment)
            for ix in range(len(rule_bases)):
                if rule_bases.fuzzy_type() == fs.FUZZY_SETS.t2:
                    aux.append(np.mean(rule_bases[ix].compute_rule_antecedent_memberships(X), axis=2) * np.expand_dims(actual_moment, axis=1))
                elif rule_bases.fuzzy_type() == fs.FUZZY_SETS.gt2:
                    aux.append(np.mean(rule_bases[ix].compute_rule_antecedent_memberships(X), axis=2) * np.expand_dims(actual_moment, axis=1))
                else:
                    aux.append(rule_bases[ix].compute_rule_antecedent_memberships(X) * np.expand_dims(actual_moment, axis=1))

        # Firing strengths shape: samples x rules
        return np.concatenate(aux, axis=1)


    def winning_rule_predict(self, X: np.array, time_moments: int) -> np.array:
        '''
        Returns the winning rule for each sample. Takes into account dominance scores if already computed.

        :param X: array with the values of the inputs.
        :return: array with the winning rule for each sample.
        '''
        consequents = []
        for ix, mrb in enumerate(self.time_mrule_bases):
            for jx, rb in enumerate(mrb):
                for rule in rb:
                    consequents.append(jx)

        # consequents = sum([[ix]*len(self[ix].get_rules())
        #                  for ix in range(len(self.rule_bases))], [])  # The sum is for flatenning
        firing_strengths = self.compute_firing_strenghts(X, time_moments)

        if self.time_mrule_bases[0].fuzzy_type() == fs.FUZZY_SETS.t2 or self.time_mrule_bases[0].fuzzy_type() == fs.FUZZY_SETS.gt2:
            association_degrees = np.mean(self.get_scores(), axis=1) * firing_strengths
        else:
            association_degrees = self.get_scores() * firing_strengths


        winning_rules = np.argmax(association_degrees, axis=1)

        return np.array([consequents[ix] for ix in winning_rules])


    def add_rule_base(self, rule_base: rl.RuleBase, time: int) -> None:
        '''
        Adds a rule base to the list of rule bases.

        :param rule_base: rule base to add.
        '''
        self.time_mrule_bases[time].add_rule_base(rule_base)


    def print_rules(self, return_rules=False) -> None:
        '''
        Print all the rules for all the consequents.
        '''
        res = ''
        for zx, time in enumerate(self.time_mrule_bases):
            res += 'Rules for time step: ' + self.time_step_names[zx] + '\n'
            res += '----------------\n' 
            for ix, ruleBase in enumerate(time):
                res += 'Consequent: ' + time.consequent_names[ix] + '\n'
                res += ruleBase.print_rules(True)
                res += '\n'
            res += '----------------\n'

        if return_rules:
            return res
        else:
            print(res)


    def get_rules(self) -> list[rl.RuleSimple]:
        '''
        Returns a list with all the rules.

        :return: list with all the rules.
        '''
        return [rule for mruleBase in self.rule_bases for rule in mruleBase.get_rules()]


    def fuzzy_type(self) -> fs.FUZZY_SETS:
        '''
        Returns the correspoing type of the RuleBase using the enum type in the fuzzy_sets module.

        :return: the corresponding fuzzy set type of the RuleBase.
        '''
        return self.time_mrule_bases[0].rule_bases[0].fuzzy_type()


    def purge_rules(self, tolerance=0.001) -> None:
        '''
        Delete the roles with a dominance score lower than the tolerance.

        :param tolerance: tolerance to delete the rules.
        '''
        for mruleBase in self.time_mrule_bases:
            mruleBase.purge_rules(tolerance)


    def __getitem__(self, item: int) -> rl.RuleBase:
        '''
        Returns the corresponding time rulebase.

        :param item: index of the rulebase.
        :return: the corresponding rulebase.
        '''
        return self.time_mrule_bases[item]


    def __len__(self) -> int:
        '''
        Returns the number of rule bases.
        '''
        return len(self.time_mrule_bases)


    def get_consequents(self) -> list[int]:
        '''
        Returns a list with the consequents of each rule base.

        :return: list with the consequents of each rule base.
        '''
        return sum([x.get_consequents() for ix, x in enumerate(self.time_mrule_bases)], [])
    

    def get_rulebases(self) -> list[rl.RuleBase]:
        '''
        Returns a list with all the rules.

        :return: list
        '''
        rule_bases = []

        for ruleBase in self.time_mrule_bases:
            for rule in ruleBase:
                rule_bases.append(rule)
        
        return rule_bases
    

    def _winning_rules(self, X: np.array, temporal_moments: list[int]) -> np.array:
        '''
        Returns the winning rule for each sample. Takes into account dominance scores if already computed.

        :param X: array with the values of the inputs.
        :return: array with the winning rule for each sample.
        '''
        
        firing_strengths = self.compute_firing_strenghts(X, temporal_moments)

        association_degrees = self.get_scores() * firing_strengths

        if (self[0].fuzzy_type() == fs.FUZZY_SETS.t2) or (self[0].fuzzy_type() == fs.FUZZY_SETS.gt2):
            association_degrees = np.mean(association_degrees, axis=2)
        elif self[0].fuzzy_type() == fs.FUZZY_SETS.gt2:
            association_degrees = np.mean(association_degrees, axis=3)

        winning_rules = np.argmax(association_degrees, axis=1)

        return winning_rules
        

#### DEFINE THE FUZZY CLASSIFIER USING TEMPORAL FUZZY SETS ####
class TemporalFuzzyRulesClassifier(evf.BaseFuzzyRulesClassifier):
    '''
    Class that is used as a classifier for a fuzzy rule based system with time dependencies. Supports precomputed and optimization of the linguistic variables.
    '''

    def __init__(self,  nRules: int = 30, nAnts: int = 4, fuzzy_type: fs.FUZZY_SETS = None, tolerance: float = 0.0,
                 n_linguist_variables: int = 0, verbose=False, linguistic_variables: list[fs.fuzzyVariable] = None,
                 domain: list[float] = None, n_class: int=None, precomputed_rules: rl.MasterRuleBase =None, runner: int=1) -> None:
        '''
        Inits the optimizer with the corresponding parameters.

        :param nRules: number of rules to optimize.
        :param nAnts: max number of antecedents to use.
        :param fuzzy type: FUZZY_SET enum type in fuzzy_sets module. The kind of fuzzy set used.
        :param tolerance: tolerance for the dominance score of the rules.
        :param n_linguist_variables: number of linguistic variables per antecedent.
        :param verbose: if True, prints the progress of the optimization.
        :param linguistic_variables: list of fuzzyVariables type. If None (default) the optimization process will init+optimize them.
        :param domain: list of the limits for each variable. If None (default) the classifier will compute them empirically.

        kwargs:
            - n_classes: number of classes to predict. Default deduces from data.
        '''
        super().__init__(nRules=nRules, nAnts=nAnts, fuzzy_type=fuzzy_type, tolerance=tolerance, 
                         n_linguist_variables=n_linguist_variables, verbose=verbose, linguistic_variables=linguistic_variables, 
                         domain=domain, n_class=n_class, precomputed_rules=precomputed_rules, runner=runner)
    

    def _contruct_tempRuleBase(self, problems, best_individuals):
        ruleBase_temp = []

        for problem, subject in zip(problems, best_individuals):
            ruleBase_time_moment = problem._construct_ruleBase(subject, 
                                                                self.fuzzy_type)
            ruleBase_temp.append(ruleBase_time_moment)

        return ruleBase_temp


    def _fix_time(self, lvs: list[temporalFuzzyVariable], time: int) -> None:
        '''
        Fixes the time of the temporal fuzzy variables.

        :param lvs: list of temporal fuzzy variables.
        :param time: integer. Time moment to fix.
        :return: None. The time is fixed.
        '''
        for lv in lvs:
            lv.fix_time(time)


    def fit(self, X: np.array, y: np.array, n_gen:int=50, pop_size:int=10, time_moments: np.array=None, checkpoints:int=0):
        '''
        Fits a fuzzy rule based classifier using a genetic algorithm to the given data.

        :param X: numpy array samples x features
        :param y: labels. integer array samples (x 1)
        :param n_gen: integer. Number of generations to run the genetic algorithm.
        :param pop_size: integer. Population size for each gneration.
        :param time_moments: array of integers. Time moments associated to each sample (when temporal dependencies are present)
        :return: None. The classifier is fitted to the data.
        '''
        problems = []
        for ix in range(len(np.unique(time_moments))):
            X_problem = X[time_moments == ix]
            y_problem = y[time_moments == ix]
            
            if self.lvs is None:
                # If Fuzzy variables need to be optimized.
                problem = evf.FitRuleBase(X_problem, y_problem, nRules=self.nRules, nAnts=self.nAnts, tolerance=self.tolerance,
                                    n_linguist_variables=self.n_linguist_variables, fuzzy_type=self.fuzzy_type, domain=self.domain, 
                                    n_classes=self.n_class, thread_runner=self.thread_runner)
            else:
                import copy
                time_lvs = [copy.deepcopy(aux) for aux in self.lvs]
                self._fix_time(time_lvs, ix)
                # If Fuzzy variables are already precomputed.       
                problem = evf.FitRuleBase(X_problem, y_problem, nRules=self.nRules, nAnts=self.nAnts,
                                    linguistic_variables=time_lvs, domain=self.domain, tolerance=self.tolerance, 
                                    n_classes=self.classes_, thread_runner=self.thread_runner)
            
            problems.append(problem)

        

        best_individuals = []
        self.performance = {}
        for time, problem in enumerate(problems):
            algorithm = GA(
            pop_size=pop_size,
            sampling=IntegerRandomSampling(),
            crossover=SBX(prob=.3, eta=3.0),
            mutation=PolynomialMutation(eta=7.0),
            eliminate_duplicates=True)

            if checkpoints > 0:
                if self.verbose:
                    print('=================================================')
                    print('n_gen  |  n_eval  |     f_avg     |     f_min    ')
                    print('=================================================')
                algorithm.setup(problem, seed=33, termination=('n_gen', n_gen)) # 33? Soon...
                for k in range(n_gen):
                    algorithm.next()
                    res = algorithm
                    if self.verbose:
                        print('%-6s | %-8s | %-8s | %-8s' % (res.n_gen, res.evaluator.n_eval, round(res.pop.get('F').mean(), 8), round(res.pop.get('F').min(), 8)))
                    if k % checkpoints == 0:
                        with open("checkpoint_" + str(time) + '_' + str(algorithm.n_gen), "w") as f:
                            pop = algorithm.pop
                            fitness_last_gen = pop.get('F')
                            best_solution_arg = np.argmin(fitness_last_gen)
                            best_individual = pop.get('X')[best_solution_arg, :]

                            rule_base = problem._construct_ruleBase(
                                best_individual, self.fuzzy_type)
                            eval_performance = evr.evalRuleBase(
                                rule_base, np.array(X), y)
                            
                            eval_performance.add_full_evaluation()  
                            # self.rename_fuzzy_variables() This wont work on checkpoints!
                            rule_base.purge_rules(self.tolerance)
                            checkpoint_rules = rule_base.print_rules(True)
                            f.write(checkpoint_rules)     

            else:
                res = minimize(problem,
                            algorithm,
                            # termination,
                            ("n_gen", n_gen),
                            copy_algorithm=False,
                            save_history=False,
                            verbose=self.verbose)

            pop = res.pop
            fitness_last_gen = pop.get('F')
            best_solution = np.argmin(fitness_last_gen)
            best_individual = pop.get('X')[best_solution, :]
            best_individuals.append(best_individual)

            if self.verbose:
                print('Rule based fit for time ' + str(time) + ' completed.')

        
            self.performance[time] = 1 - fitness_last_gen[best_solution]

        try:
            self.var_names = list(X.columns)
            self.X = X.values
        except AttributeError:
            self.X = X
            self.var_names = [str(ix) for ix in range(X.shape[1])]

        
        ruleBase_temp = self._contruct_tempRuleBase(problems, best_individuals)

        self.rule_base = temporalMasterRuleBase(ruleBase_temp)

        self.eval_performance = evr.evalRuleBase(self.rule_base, np.array(X), y, time_moments)

        self.eval_performance.add_full_evaluation()  
        self.rename_fuzzy_variables()
        self.rule_base.purge_rules(self.tolerance)


    def forward(self, X: np.array, time_moments: list[int] = None) -> np.array:
        '''
        Returns the predicted class for each sample.

        :param X: np array samples x features.
        :param time_moments: list of integers. Time moments associated to each sample (when temporal dependencies are present)
        :return: np array samples (x 1) with the predicted class.
        '''
        try:
            X = X.values  # If X was a numpy array
        except AttributeError:
            pass
        
        return self.rule_base.winning_rule_predict(X, time_moments)



    def plot_fuzzy_variables(self) -> None:
        '''
        Plot the fuzzy partitions in each fuzzy variable.
        '''
        fuzzy_variables = self.rule_base.rule_bases[0].antecedents

        for ix, fv in enumerate(fuzzy_variables):
            vis_rules.plot_fuzzy_variable(fv)


    def get_rulebase(self) -> list[np.array]:
        '''
        Get the rulebase obtained after fitting the classifier to the data.

        :return: a matrix format for the rulebase.
        '''
        return self.rule_base.get_rulebase_matrix()


def eval_temporal_fuzzy_model(fl_classifier: evf.BaseFuzzyRulesClassifier, X_train: np.array, y_train: np.array,
                     X_test: np.array, y_test: np.array, time_moments: list[int] = None, test_time_moments: list[int] = None,
                     plot_rules=True, print_rules=True, plot_partitions=True, return_rules=False, print_accuracy=True, print_matthew=True) -> None:
    '''
    Function that evaluates a fuzzy rule based model. It also plots the rules and the fuzzy partitions.

    :param fl_classifier: Fuzzy rule based model.
    :param X_train: Training data.
    :param y_train: Training labels.
    :param X_test: Test data.
    :param y_test: Test labels.
    :param plot_rules: If True, it plots the rules.
    :param print_rules: If True, it prints the rules.
    :param plot_partitions: If True, it plots the fuzzy partitions.
    :return: None
    '''

    if print_accuracy:
      print('ACCURACY')
      print('Train performance: ' +
            str(np.mean(np.equal(y_train, fl_classifier.forward(X_train, time_moments)))))
      print('Test performance: ' +
            str(np.mean(np.equal(y_test, fl_classifier.forward(X_test, test_time_moments)))))
      print('------------')
    if print_matthew:
      print('MATTHEW CORRCOEF')
      print('Train performance: ' +
            str(matthews_corrcoef(y_train, fl_classifier.forward(X_train, time_moments))))
      print('Test performance: ' +
            str(matthews_corrcoef(y_test, fl_classifier.forward(X_test, test_time_moments))))
      print('------------')

    for ix in np.unique(time_moments):
      try:
            X_aux = X_train[time_moments == ix, :]
            X_aux_test = X_test[time_moments == ix, :]
      except pd.core.indexing.InvalidIndexError:
            X_aux = X_train.iloc[time_moments == ix, :]
            X_aux_test = X_test.iloc[test_time_moments == ix, :]

      y_aux = y_train[time_moments == ix]
      y_aux_test = y_test[test_time_moments == ix]

      if print_matthew:
            print('MOMENT ' + str(ix))
            print('------------')
            print('MATTHEW CORRCOEF')
            print('Train performance: ' +
                  str(matthews_corrcoef(y_aux, fl_classifier.forward(X_aux, np.array([ix] * X_aux.shape[0])))))
            print('Test performance: ' +
                  str(matthews_corrcoef(y_aux_test, fl_classifier.forward(X_aux_test, np.array([ix] * X_aux_test.shape[0])))))
            print('------------')

    if plot_rules:
        vis_rules.visualize_rulebase(fl_classifier)
    if print_rules or return_rules:
        res = fl_classifier.print_rules(return_rules)
    if plot_partitions:
        fl_classifier.plot_fuzzy_variables()
      
    if return_rules:
        return res
    else:
        print(res)





'''
Module to keep track of the usage of the different features of the library. 
Asks nicely for permission before saving anything.

'''
import os as _os
import atexit

from enum import Enum

usage_categories = Enum('uses', ['FuzzySets', 'Funcs', 'Persistence', 'Visualization', 'FuzzyCognitiveMaps', 'RuleMining', 'Classification'])

def instance_dict_usage():
    # Inits the categories to track
    return {usage_categories.FuzzySets: {'t1':0 , 't2': 0, 'gt2': 0, 'temporal': 0, 'temporal_t2': 0, 'categorical': 0},
                usage_categories.Funcs: {'fit': 0, 'precompute_labels': 0, 'opt_labels':0},
                usage_categories.Persistence : {'persistence_write': 0,
                    'persistence_read': 0},
                usage_categories.FuzzyCognitiveMaps : {'fcm_create': 0, 'fcm_report': 0},
                usage_categories.Visualization : {'plot_rules': 0, 'plot_graph': 0, 'print_rules': 0, 'plot_fuzzy_variable': 0},
                usage_categories.RuleMining : {'mine_rulebase' : 0},
                usage_categories.Classification : {'double_go': 0, 'data_mining': 0}
    }


def rename_dict_usage_categories():
    global usage_data
    str_names = ['Fuzzy sets', 'Utils and training', 'Persistence', 'Fuzzy Cognitive Maps', 'Visualization', 'Rule Mining', 'Classification']
    new_usage = {}

    for ix, (key, value) in enumerate(usage_data.items()):
        new_usage[str_names[ix]] = value

    usage_data = new_usage


def send_data():
    global path_usage_data_folder
    # Send the data to the developers using a sfpt server (TODO)
    import paramiko

    host = "sftp.upv.es"
    port = 22

    files_to_send = _os.listdir(path_usage_data_folder)

    # create ssh client 
    ssh_client = paramiko.SSHClient()

    # remote server credentials
    host = "hostname"
    # username = "username"
    # password = "password"
    port = port

    ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh_client.connect(hostname=host,port=port)

    # create an SFTP client object
    ftp = ssh_client.open_sftp()

    for file in files_to_send:
        # send a file from the remote server
        ftp.put(path_usage_data_folder + '/' + file, '/home/' + file)
    
    # Send also the logger file
    ftp.put('ex_fuzzy.log', '/home/ex_fuzzy.log')
    
    # close the connection
    ftp.close()
    ssh_client.close()

    # Clear the logger file
    with open('ex_fuzzy.log', 'w') as f:
        f.write('')


def main():
    global path_usage_data_folder
    # Open the usage file from user
    file_name = 'usage_data_permission.txt'
    directory = _os.path.dirname(_os.path.realpath(__file__))

    if file_name not in _os.listdir(directory):
        # Get input from user
        print('exFuzzy:')
        print('Do you want to share your usage data with the developers?')
        print('This data will be used to improve the library. It will also help us to acquire funding for the project.')
        print('If you agree, we will store the number of times you use each functionality in a file called "usage_data.txt".')
        print('This file will be stored in the same folder as the library.')
        print('If you do not agree, we will not store any data.')
        print('If you agree, type "[y]es". If you do not agree, type "no".')
        user_input = None

        while user_input not in ['no', 'yes', 'y']:
            user_input = input()

            if user_input not in ['no', 'yes', 'y']:
                print('Answer not understood, please repeat: If you agree, type "[y]es". If you do not agree, type "no".')
        
        with open(_os.path.join(directory, file_name), 'w') as f:
            f.write(user_input)

    # Read the usage file
    with open(_os.path.join(directory, file_name), 'r') as f:
        user_input = f.read()

    save_usage_flag = user_input == 'yes' or user_input == 'y'

    if save_usage_flag:
        import logging
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG)
        # Create handlers
        c_handler = logging.StreamHandler()
        f_handler = logging.FileHandler(_os.path.join(directory, 'ex_fuzzy.log'))
        c_handler.setLevel(logging.WARNING)
        f_handler.setLevel(logging.DEBUG)
        # Create formatters and add it to handlers
        c_format = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
        f_format = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        c_handler.setFormatter(c_format)
        f_handler.setFormatter(f_format)
        # Add handlers to the logger
        logger.addHandler(c_handler)
        logger.addHandler(f_handler)

        # Create usage reports
        @atexit.register
        def save_usage():
            # Saves the usage in the corresponding file
            path_usage_data = _os.path.join(path_usage_data_folder, 'usage_data_' + str(number_files) + '.txt')
            rename_dict_usage_categories()
            import json
            # Save the usage as a json
            with open(_os.path.join(directory, path_usage_data), 'w') as f:
                json.dump(usage_data, f)
            
            # Send the data to the developers
            # send_data()
        # Create usage data dictionary
        usage_data = instance_dict_usage()

        path_usage_data_folder = 'usage_data'
        if path_usage_data_folder not in _os.listdir(directory):
            _os.mkdir(_os.path.join(directory, path_usage_data_folder))
        
        number_files = len(_os.listdir(_os.path.join(directory, path_usage_data_folder)))

        if number_files > 100:
            # Avoid a large number of files
            while number_files > 100:
                _os.remove(_os.path.join(directory, path_usage_data_folder, 'usage_data_' + str(number_files-1) + '.txt'))
                number_files = len(_os.listdir(path_usage_data_folder))


        atexit.register(save_usage)

save_usage_flag = False
if __name__ == '__main__':
    pass


import numpy as np
import math
import sys

sys.path.append('./ex_fuzzy/')
sys.path.append('../ex_fuzzy/')
import ex_fuzzy as ex_fuzzy

sample_size = 10000
tolerate = 0.05
def test_quartiles():
    sample = np.random.random_sample(sample_size)
    targets = [0, 0.25, 0.5, 1]
    quartiles = ex_fuzzy.utils.quartile_compute(sample)
    for ix, target in enumerate(targets):
        assert math.isclose(quartiles[ix], target, abs_tol=tolerate), 'Not the correct ' +str(target) + ' quartile.'

def test_quantiles():
    sample = np.random.random_sample(sample_size)
    targets = [0, 0.20, 0.30, 0.45, 0.55, 0.7, 0.8, 1]
    quartiles = ex_fuzzy.utils.fixed_quantile_compute(sample)
    for ix, target in enumerate(targets):
        assert math.isclose(quartiles[ix], target, abs_tol=tolerate), 'Not the correct ' +str(target) + ' quartile.'

def test_3_partitions():
    sample = np.random.random_sample(sample_size)
    targets = [0, 0.20, 0.50, 0.80, 1.00]
    quartiles = ex_fuzzy.utils.partition3_quantile_compute(sample)
    for ix, target in enumerate(targets):
        assert math.isclose(quartiles[ix], target, abs_tol=tolerate), 'Not the correct ' +str(target) + ' quartile.'


def test_construct_partitions_t1():
    sample = np.random.random_sample((sample_size, 1))
    partitions = ex_fuzzy.utils.construct_partitions(sample, ex_fuzzy.fuzzy_sets.FUZZY_SETS.t1)
    assert len(partitions[0]) == 3, 'Not the correct number of partitions'
    assert math.isclose(partitions[0](0.01)[0], 1, abs_tol=tolerate), 'Not the correct partition'


def test_construct_partitions_t2():
    sample = np.random.random_sample((sample_size, 1))
    partitions = ex_fuzzy.utils.construct_partitions(sample, ex_fuzzy.fuzzy_sets.FUZZY_SETS.t2)
    assert len(partitions[0]) == 3, 'Not the correct number of partitions'
    assert math.isclose(partitions[0](0.01)[0][0], 0.8, abs_tol=tolerate), 'Not the correct partition'
    assert math.isclose(partitions[0](0.01)[0][1], 1, abs_tol=tolerate), 'Not the correct partition'


def test_construct_partitions_gt2():
    sample = np.random.random_sample((sample_size, 1))
    partitions = ex_fuzzy.utils.construct_partitions(sample, ex_fuzzy.fuzzy_sets.FUZZY_SETS.gt2)
    assert len(partitions[0]) == 3, 'Not the correct number of partitions'
    assert math.isclose(np.mean(partitions[0][0].alpha_reduction(partitions[0](0.1)[0])), 0.9, abs_tol=tolerate), 'Not the correct partition'


def test_temporal_conditional():
    sample_size = 10
    sample = np.random.random_sample((sample_size, ))

    first_temp = int(sample_size / 2)

    sample[:first_temp] = sample[:first_temp] * 0.5
    time_labels = [0] * sample_size
    time_labels[first_temp:] = [1] * first_temp

    vl1 = ex_fuzzy.fuzzy_sets.FS('',[0, 0.20, 0.30, 0.5], [0, 1])
    vl2 = ex_fuzzy.fuzzy_sets.FS('',[0.5, 0.70, 0.80, 1], [0, 1])

    conditional_frequencies = ex_fuzzy.utils.construct_conditional_frequencies(sample, time_labels, [vl1, vl2])

    assert conditional_frequencies[0, 0] > conditional_frequencies[0, 1]
    assert conditional_frequencies[1, 0] < conditional_frequencies[1, 1]


def test_temporal_assemble():
    sample_size = 1000
    sample = np.random.random_sample((sample_size, 5))
    y = np.random.randint(0, 2, sample_size)

    temp_moments1 = np.random.randint(0, 2, sample_size)
    temp_moments2 = 1 - temp_moments1
    temp_moments = [temp_moments1, temp_moments2]

    [X_train, X_test, y_train, y_test], [train_temporal_boolean_markers, test_temporal_boolean_markers] = ex_fuzzy.utils.temporal_assemble(sample, y, temp_moments)

    t1test = [y_test[jx] for jx, aux in enumerate(test_temporal_boolean_markers[0]) if aux]
    t2test = [y_test[jx] for jx, aux in enumerate(test_temporal_boolean_markers[1]) if aux]

    assert len(t1test) == len(t2test)

    


if __name__ == '__main__':
    test_quartiles()
    test_quantiles()
    test_3_partitions()
    test_construct_partitions_t1()
    test_construct_partitions_t2()
    test_construct_partitions_gt2()
    test_temporal_conditional()
    test_temporal_assemble()
    print('All tests passed.')


from __future__ import annotations
'''
Module to use fuzzy cognitive maps.

The module contains the class FuzzyCognitiveMap, which is used to create and
use fuzzy cognitive maps. You can also plot them, or simulate the dynamics of
the FCM.

For the original papers about FCM, see the works by Bart Kosko.

'''
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

try:
    from . import maintenance as mnt
except ImportError:

    import maintenance as mnt


def _threshold_modules(connections: np.array | pd.DataFrame, threshold) -> np.array | pd.DataFrame:
    '''Thresholds the connections matrix to the {-1, 0, 1} values.'''
    return np.abs(connections) > threshold * np.sign(connections)


def __look_periods(states: list[np.array], min_period_len=2) -> list[np.array]:
    '''Looks for periods in the states list. Returns the period if found, None otherwise.'''
    max_period_len = int(len(states) / 2)

    for period_len in np.arange(max_period_len, min_period_len, -1):

        for i in range(len(states)):
            candidate = states[i:i+period_len]
            next_candidate = states[i+period_len:i+period_len*2]
            if len(next_candidate) < min_period_len:
                break
            if candidate != next_candidate:
                break
        
        return candidate


    return None


def __look_attractors(states: list[np.array]) -> [bool, np.array]:
    '''Checks if all the states in the list are the same'''
    attractors = []
    for state in states:
        if not state in attractors:
            attractors.append(state)
        else:
            return False, []
    
    return True, attractors[0]
    

def look_pattern_states(fcm: FuzzyCognitiveMap, sim_steps: int, pattern_len: 50, max_period_size: 50) -> [np.array]:
    '''Looks for the pattern states of the FCM when simulation is prolongued.
    
    
    :param fcm : FuzzyCognitiveMap. The FCM to look for the attractor states.
    :param max_steps: int. The maximum number of steps to look for the attractor states.
    :param random_inits : int
    :returns: list of np.array. The attractor states found. None if none were found
    '''
    for _ in range(sim_steps):
        state = fcm.step()
    
    steps = []
    for _ in range(pattern_len):
        state = fcm.step()
        steps.append(state)

    satisfactory, period = __look_attractors(steps)
    if not satisfactory:
        period = __look_periods(steps, min_period_len=2)
        
    return period


def study_attractors_FCM(fcm: FuzzyCognitiveMap, max_steps: int, random_inits: int=10) -> [np.array]:
    '''Looks for the attractor states of the FCM when simulation is prolongued.
    
    
    :param fcm : FuzzyCognitiveMap. The FCM to look for the attractor states.
    :param max_steps: int. The maximum number of steps to look for the attractor states.
    :param random_inits : int
    :returns: list of np.array. The attractor states found. None if none were found
    '''
    if mnt.save_usage_flag:
        mnt.usage_data[mnt.usage_categories.FuzzyCognitiveMaps]['fcm_report'] += 1

    attractors = {}
    gen_random_state = lambda : np.random.randint(0, 2, fcm.connections.shape[0])
    for _ in range(random_inits):
        init_state = gen_random_state()
        fcm.set_state(init_state)
        attractors[init_state] = look_pattern_states(fcm, max_steps)
    
    return attractors


def attractors_report(attractors: dict[np.array, np.array]) -> None:
    '''
    Prints a report of the attractors found.
    
    :param attractors: dict[np.array, np.array]. The attractors found.
    '''
    if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.FuzzyCognitiveMaps]['fcm_report'] += 1
    pattern_dict = {}

    for _, attractor in attractors.items():
        if attractor is None:
            pattern_dict['Chaotic'] = pattern_dict.get('Chaotic', 0) + 1 / len(attractors)
        else:
            pattern_dict[str(attractor)] = pattern_dict.get(str(attractor), 0) + 1 / len(attractors)

    state_dict = {}
    list_states = []
    for _, attractor in attractors.items():
        if attractor is not None:
            for state in attractor:
                list_states += [str(state)]

    for state in list_states:
        state_dict[state] = state_dict.get(state, 0) + 1 / len(list_states)

    return pattern_dict, state_dict


class FuzzyCognitiveMap:

    def __init__(self, connections: np.array | pd.DataFrame, threshold:int=0.5) -> None:
        '''
        Creates a fuzzy cognitive map.
        
        
        :param connections: np.array | pd.DataFrame. A square matrix with the connections between the concepts.
        :param threshold: int, optional. When simulating steps the state
        will be trimmed using these threhold into the {0, 1, -1} values.
        The default is 0.5.
        '''
        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.FuzzyCognitiveMaps]['fcm_create'] += 1

        self.connections = connections
        self.state = np.zeros(connections.shape[0])
        self.threshold = threshold
    

    def var_names(self) -> list[str]:
        '''Returns the names of the variables.'''
        try:
            return list(self.connections.columns)
        except AttributeError:
            return None
    

    def step(self) -> np.array | pd.DataFrame:
        '''Performs a step in the FCM given the actual state.'''
        self.state = _threshold_modules(self.state @ self.connections, self.threshold)
    
        if isinstance(self.state, pd.DataFrame):
            return pd.DataFrame(self.state, columns=self.var_names())
        else:
            return self.state


    def simulate(self, steps: int) -> np.array | pd.DataFrame:
        '''
        Simulates the FCM for a number of steps.
        
        :param steps: int. The number of steps to simulate.
        '''
        for _ in range(steps):
            fstep = self.step()
        
        return fstep
    

    def add(self, other: FuzzyCognitiveMap) -> None:
        '''Adds the connections of other FCM to the actual FCM.'''
        self.connections = self.connections + other.connections


    def set_state(self, state: np.array | pd.DataFrame) -> None:
        '''Sets the state of the FCM.'''
        try:
            self.state = state.values
        except AttributeError:
            self.state = state


    def set_and_step(self, state: np.array | pd.DataFrame) -> np.array | pd.DataFrame:
        '''Sets the state of the FCM and performs a step.'''
        self.set_state(state)
        return self.step()
    

    def set_and_simulate(self, state: np.array | pd.DataFrame, steps: int) -> np.array | pd.DataFrame:
        '''Sets the state of the FCM and performs a simulation.'''
        self.set_state(state)
        return self.simulate(steps)


    def clear_state(self) -> None:
        '''Clears the state of the FCM.'''
        self.state = np.zeros(self.connections.shape[0])


    def __add__(self, other: FuzzyCognitiveMap) -> FuzzyCognitiveMap:
        '''Creates a new FCM that is the addition of the two different connection matrix.'''
        return FuzzyCognitiveMap(self.connections + other.connections)



'''
Load the rules of a fuzzy rules system using plain text format.

'''
import numpy as np

try:
    from . import fuzzy_sets as fs
    from . import rules
    from . import maintenance as mnt

except ImportError:
    import fuzzy_sets as fs
    import rules
    import maintenance as mnt


def load_fuzzy_rules(rules_printed: str, fuzzy_variables: list) -> rules.MasterRuleBase:
    '''
    Load the rules from a string.
    
    :param rules_printed: string with the rules. Follows the specification given by the same printing method of rules.MasterRuleBase
    :param fuzzy_variables: list with the linguistic variables. Objects of FuzzyVariable class.
    
    '''
    if mnt.save_usage_flag:
        mnt.usage_data[mnt.usage_categories.Persistence]['persistence_read'] += 1
        
    consequent = 0
    linguistic_variables_names = [linguistic_variable.name for linguistic_variable in fuzzy_variables]
    value_names = [x.name for x in fuzzy_variables[0]]
    fz_type = fuzzy_variables[0].fuzzy_type()
    consequent_names = []
    for line in rules_printed.splitlines():
        if line.startswith('IF'):
            #Is a rule
            antecedents , consequent_ds = line.split('WITH')
            consequent_ds = consequent_ds.split(',')[0].strip()
            init_rule_antecedents = np.zeros(
                (len(fuzzy_variables),)) - 1  # -1 is dont care
            
            for antecedent in antecedents.split('AND'):
                antecedent = antecedent.replace('IF', '').strip()
                antecedent_name, antecedent_value = antecedent.split('IS')
                antecedent_name = antecedent_name.strip()
                antecedent_value = antecedent_value.strip()
                antecedent_index = linguistic_variables_names.index(antecedent_name)
                antecedent_value_index = value_names.index(antecedent_value)

                init_rule_antecedents[antecedent_index] = antecedent_value_index
                
            rule_simple = rules.RuleSimple(init_rule_antecedents, 0)
            rule_simple.score = float(consequent_ds[3:].strip()) # We remove the 'DS ' and the last space
            reconstructed_rules.append(rule_simple)

        elif line.startswith('Rules'):
            #New consequent
            consequent_name = line.split(':')[-1].strip()
            consequent_names.append(consequent_name)
            if consequent > 0:
                if fz_type == fs.FUZZY_SETS.t1:
                    rule_base = rules.RuleBaseT1(fuzzy_variables, reconstructed_rules)
                elif fz_type == fs.FUZZY_SETS.t2:
                    rule_base = rules.RuleBaseT2(fuzzy_variables, reconstructed_rules)
                elif fz_type == fs.FUZZY_SETS.gt2:
                    rule_base = rules.RuleBaseGT2(fuzzy_variables, reconstructed_rules)
                                    
            if consequent == 1:
                mrule_base = rules.MasterRuleBase([rule_base])
            elif consequent > 1:
                mrule_base.add_rule_base(rule_base)

            reconstructed_rules = []
            consequent += 1
    
    # We add the last rule base
    if fz_type == fs.FUZZY_SETS.t1:
        rule_base = rules.RuleBaseT1(fuzzy_variables, reconstructed_rules)
    elif fz_type == fs.FUZZY_SETS.t2:
        rule_base = rules.RuleBaseT2(fuzzy_variables, reconstructed_rules)
    elif fz_type == fs.FUZZY_SETS.gt2:
        rule_base = rules.RuleBaseGT2(fuzzy_variables, reconstructed_rules)
        
    mrule_base.add_rule_base(rule_base) 
    mrule_base.rename_cons(consequent_names)

    return mrule_base




"""
This is a the source file that contains the class of GT2 fuzzy set and its most direct applications, 
like computing the FM function, etc.

"""
import enum
from typing import Generator

import numpy as np

try:
    from . import maintenance as mnt
except:
    import maintenance as mnt

# You dont require torch to use this module, however, we need to import it to give support in case you feed these methods with torch tensors.
try:
    import torch
    torch_available = True
except:
    torch_available = False


''' Enum that defines the fuzzy set types.'''
class FUZZY_SETS(enum.Enum):
    t1 = 'Type 1'
    t2 = 'Type 2'
    gt2 = 'General Type 2'


    def __eq__(self, __value: object) -> bool:
        return self.value == __value.value


def trapezoidal_membership(x: np.array, params: list[float], epsilon=10E-5) -> np.array:
    '''
    Trapezoidal membership functions.

    :param x: input values in the fuzzy set referencial domain.
    :param params:  four numbers that comprises the start and end of the trapezoid.
    :param epsilon: small float number for numerical stability. Adjust accordingly only if there are NaN issues.
    '''
    a, b, c, d = params

    # Special case: a singleton trapezoid
    if a == d:
        # If they are numpy arrays, we need to use the numpy function
        if isinstance(x, np.ndarray):
            return np.equal(x, a).astype(float)
        if torch_available:
            if isinstance(x, torch.Tensor):
                return torch.eq(x, a).float()
            

    if b == a:
        b += epsilon
    if c == d:
        d += epsilon

    aux1 = (x - a) / (b - a)
    aux2 = (d - x) / (d - c)

    if isinstance(x, np.ndarray):
        return np.clip(np.minimum(aux1, aux2), 0.0, 1.0)
    elif isinstance(x, torch.Tensor):
        return torch.clamp(torch.min(aux1, aux2), 0.0, 1.0)


def __gaussian2(x, params: list[float]) -> np.array:
    '''
    Gaussian membership functions.

    :param mean:  real number, mean of the gaussian function.
    :param amplitude: real number.
    :param standard_deviation: std of the gaussian function.
    '''
    mean, amplitude, standard_deviation = params
    return amplitude * np.exp(- ((x - mean) / standard_deviation) ** 2)


class FS():
    '''
    Class that defines the most basic fuzzy sets (also known as Type 1 fuzzy sets or Zadeh sets).
    '''

    def __init__(self, name: str, membership_parameters: list[float], domain: list[float]) -> None:
        '''
        Creates a fuzzy set.

        :param name: string.
        :param secondMF_lower: four real numbers. Parameters of the lower trapezoid/gaussian function.
        :param secondMF_upper: four real numbers. Parameters of the upper trapezoid/gaussian function.
        :param domain: list of two numbers. Limits of the domain of the fuzzy set.
        '''
        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.FuzzySets][self.type().name] += 1
            
        self.name = name
        self.domain = domain
        self.membership_parameters = membership_parameters


    def membership(self, x: np.array) -> np.array:
        '''
        Computes the membership of a point or a vector.

        :param x: input values in the fuzzy set referencial domain.
        '''
        return trapezoidal_membership(x, self.membership_parameters)


    def type(self) -> FUZZY_SETS:
        '''
        Returns the corresponding fuzzy set type according to FUZZY_SETS enum.

        :return: FUZZY_SETS enum.
        '''
        return FUZZY_SETS.t1


    def __call__(self,  x: np.array) -> np.array:
        '''
        Calling the Fuzzy set returns its membership.

        :param x: input values in the fuzzy set referencial domain.
        :return: membership of the fuzzy set.
        '''
        return self.membership(x)


class categoricalFS(FS):

    def __init__(self, name: str, category) -> None:
        '''
        Creates a categorical fuzzy set. It gives 1 to the category and 0 to the rest.
        Use it when the variable is categorical and the categories are known, so that rule inference systems
        can naturally support both crisp and fuzzy variables.

        :param name: string.
        :param categories: list of strings. Possible categories.
        '''
        self.name = name
        self.category = category

        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.FuzzySets][self.type().name] += 1


    def membership(self, x: np.array) -> np.array:
        '''
        Computes the membership of a point or vector of elements.

        :param x: input values in the referencial domain.
        '''
        if isinstance(x, np.ndarray):
            res = np.equal(x, self.category).astype(float)
        if torch_available:
            if isinstance(x, torch.Tensor):
                res = torch.eq(x, self.category).float()

        return res


    def type(self) -> FUZZY_SETS:
        '''
        Returns the corresponding fuzzy set type according to FUZZY_SETS enum.
        '''
        return FUZZY_SETS.t1


class IVFS(FS):
    '''
    Class to define a iv fuzzy set.
    '''

    def __init__(self, name: str, secondMF_lower: list[float], secondMF_upper: list[float], 
                domain: list[float], lower_height=1.0) -> None:
        '''
        Creates a IV fuzzy set.

        :param name: string.
        :param secondMF_lower: four real numbers. Parameters of the lower trapezoid/gaussian function.
        :param secondMF_upper: four real numbers. Parameters of the upper trapezoid/gaussian function.
        :param domain: list of two numbers. Limits of the domain if the fuzzy set.
        '''
        self.name = name
        self.domain = domain

        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.FuzzySets][self.type().name] += 1

        assert secondMF_lower[0] >= secondMF_upper[0], 'First term membership incoherent'
        assert secondMF_lower[0] <= secondMF_lower[1] and secondMF_lower[
            1] <= secondMF_lower[2] and secondMF_lower[2] <= secondMF_lower[3], 'Lower memberships incoherent. '
        assert secondMF_upper[0] <= secondMF_upper[1] and secondMF_upper[
            1] <= secondMF_upper[2] and secondMF_upper[2] <= secondMF_upper[3], 'Upper memberships incoherent.'
        assert secondMF_lower[3] <= secondMF_upper[3], 'Final term memberships incoherent.'

        self.secondMF_lower = secondMF_lower
        self.secondMF_upper = secondMF_upper
        self.lower_height = lower_height


    def membership(self, x: np.array) -> np.array:
        '''
        Computes the iv-membership of a point or a vector.

        :param x: input values in the fuzzy set referencial domain.
        :return: iv-membership of the fuzzy set.
        '''
        lower = trapezoidal_membership(
            x, self.secondMF_lower) * self.lower_height
        upper = trapezoidal_membership(x, self.secondMF_upper)

        try:
            assert np.all(lower <= upper)
        except AssertionError:
            np.argwhere(lower > upper)

        return np.stack([lower, upper], axis=-1)


    def type(self) -> FUZZY_SETS:
        '''
        Returns the corresponding fuzzy set type according to FUZZY_SETS enum: (t2)
        '''
        return FUZZY_SETS.t2


class categoricalIVFS(IVFS):
    '''
    Class to define a iv fuzzy set with categorical membership.
    '''

    def __init__(self, name: str, category) -> None:
        '''
        Creates a categorical iv fuzzy set. It gives 1 to the category and 0 to the rest.
        Use it when the variable is categorical and the categories are known, so that rule inference systems
        can naturally support both crisp and fuzzy variables.

        :param name: string.
        :param categories: list of strings. Possible categories.
        :param domain: list of two numbers. Limits of the domain if the fuzzy set.
        '''
        self.name = name
        self.category = category

        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.FuzzySets][self.type().name] += 1


    def membership(self, x: np.array) -> np.array:
        '''
        Computes the membership of a point or vector of elements.

        :param x: input values in the referencial domain.
        '''
        res = np.equal(x, self.category).astype(float)
        res = np.stack([res, res], axis=-1)
        
        return res


    def type(self) -> FUZZY_SETS:
        '''
        Returns the corresponding fuzzy set type according to FUZZY_SETS enum.
        '''
        return FUZZY_SETS.t2
    


class GT2(FS):
    '''
    Class to define a gt2 fuzzy set.
    '''

    MAX_RES_SUPPORT = 4  # Number of decimals supported in the secondary

    def __init__(self, name: str, secondary_memberships: dict[float, FS], 
                domain: list[float], significant_decimals: int, 
                alpha_cuts: list[float] = [0.2, 0.4, 0.5, 0.7, 0.9, 1.0], 
                unit_resolution: float = 0.2) -> None:
        '''
        Creates a GT2 fuzzy set.

        :param name: string.
        :param secondary_memberships: list of fuzzy sets. Secondary membership that maps original domain to [0, 1]
        :param secondMF_upper: four real numbers. Parameters of the upper trapezoid/gaussian function.
        :param domain: list of two numbers. Limits of the domain if the fuzzy set.
        :param alpha_cuts: list of real numbers. Alpha cuts of the fuzzy set.
        :param unit_resolution: real number. Resolution of the primary membership function.
        '''
        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.FuzzySets][self.type().name] += 1

        self.name = name
        self.domain = domain
        self.secondary_memberships = secondary_memberships
        self.alpha_cuts = alpha_cuts
        self.iv_secondary_memberships = {}
        self.unit_resolution = unit_resolution
        og_keys = list(self.secondary_memberships.keys())
        self.significant_decimals = significant_decimals
        self.domain_init = int(
            float(og_keys[0]) * 10**self.significant_decimals)

        formatted_keys = [('%.' + str(self.significant_decimals) + 'f') %
                          xz for xz in np.array(list(self.secondary_memberships.keys()))]
        for og_key, formatted_key in zip(og_keys, formatted_keys):
            secondary_memberships[formatted_key] = self.secondary_memberships.pop(
                og_key)

        self.sample_unit_domain = np.arange(
            0, 1 + self.unit_resolution, self.unit_resolution)

        for ix, alpha_cut in enumerate(alpha_cuts):
            level_memberships = {}
            array_level_memberships = np.zeros(
                (len(secondary_memberships.items()), 2))

            for jx, (x, fs) in enumerate(secondary_memberships.items()):
                alpha_primary_memberships = fs.membership(
                    self.sample_unit_domain)
                alpha_membership = alpha_primary_memberships >= alpha_cut

                try:
                    b = self.sample_unit_domain[np.argwhere(
                        alpha_membership)[0]][0]
                    c = self.sample_unit_domain[np.argwhere(
                        alpha_membership)[-1]][0]
                except IndexError:  # If no numbers are bigger than alpha, then it is because of rounding errors near 0. So, we fix this manually.
                    b = 0
                    c = 0

                alpha_cut_interval = [b, c]

                level_memberships[x] = alpha_cut_interval
                array_level_memberships[jx, :] = np.array(alpha_cut_interval)

            self.iv_secondary_memberships[alpha_cut] = array_level_memberships


    def membership(self, x: np.array) -> np.array:
        '''
        Computes the alpha cut memberships of a point.

        :param x: input values in the fuzzy set referencial domain.
        :return: np array samples x alpha_cuts x 2
        '''
        x = np.array(x)
        # locate the x in the secondary membership
        formatted_x = (
            x * 10**self.significant_decimals).astype(int) - self.domain_init

        # Once the x is located we compute the function for each alpha cut
        alpha_cut_memberships = []
        for ix, alpha in enumerate(self.alpha_cuts):
            ivfs = np.squeeze(
                np.array(self.iv_secondary_memberships[alpha][formatted_x]))
            alpha_cut_memberships.append(np.squeeze(ivfs))

        alpha_cut_memberships = np.array(alpha_cut_memberships)

        # So that the result is samples x alpha_cuts x 2
        return np.swapaxes(alpha_cut_memberships, 0, 1)


    def type(self) -> FUZZY_SETS:
        return FUZZY_SETS.gt2
    

    def _alpha_reduction(self, x) -> np.array:
        '''
        Computes the type reduction to reduce the alpha cuts to one value.

        :param x: array with the values of the inputs.
        :return: array with the memberships of the consequents for each sample.
        '''
        if len(x.shape) == 3:
            formatted = np.expand_dims(np.expand_dims(
                np.array(self.alpha_cuts), axis=1), axis=0)
        else:
            formatted = self.alpha_cuts
        return np.sum(formatted * x, axis=-1) / np.sum(self.alpha_cuts)
    

    def alpha_reduction(self, x) -> np.array:
        '''
        Computes the type reduction to reduce the alpha cuts to one value.

        :param x: array with the values of the inputs.
        :return: array with the memberships of the consequents for each sample.
        '''
        return self._alpha_reduction(x)
        

class gaussianIVFS(IVFS):
    '''
    Class to define a iv fuzzy set with gaussian membership.
    '''

    def membership(self, input: np.array) -> np.array:
        '''
        Computes the gaussian iv-membership of a point or a vector.

        :param input: input values in the fuzzy set referencial domain.
        :return: np array samples x 2
        '''
        lower = __gaussian2(input, self.secondMF_lower)
        upper = __gaussian2(input, self.secondMF_upper)

        return np.array(np.concatenate([lower, upper])).T


    def type(self) -> FUZZY_SETS:
        '''
        Returns the type of the fuzzy set. (t1)
        '''
        return FUZZY_SETS.t1


class fuzzyVariable():
    '''
    Class to implement a fuzzy Variable. Contains a series of fuzzy sets and computes the memberships to all of them.
    '''

    def __init__(self, name: str, fuzzy_sets: list[FS]) -> None:
        '''
        Creates a fuzzy variable.

        :param name: string. Name of the fuzzy variable.
        :param fuzzy_sets: list of IVFS. Each of the fuzzy sets that comprises the linguist variables of the fuzzy variable.
        '''
        self.linguistic_variables = []
        self.name = name

        for ix, fs in enumerate(fuzzy_sets):
            self.linguistic_variables.append(fs)

        self.fs_type = self.linguistic_variables[0].type()


    def linguistic_variable_names(self) -> list:
        '''
        Returns the name of the linguistic variables.

        :return: list of strings.
        '''
        return [fs.name for fs in self.linguistic_variables]


    def get_linguistic_variables(self) -> list[FS]:
        '''
        Returns the name of the linguistic variables.

        :return: list of strings.
        '''
        return self.linguistic_variables


    def compute_memberships(self, x: np.array) -> list:
        '''
        Computes the membership to each of the FS in the fuzzy variables.

        :param x: numeric value or array. Computes the membership to each of the FS in the fuzzy variables.
        :return: list of floats. Membership to each of the FS in the fuzzy variables.
        '''
        res = []

        for fuzzy_set in self.linguistic_variables:
            res.append(fuzzy_set.membership(x))

        return res


    def domain(self) -> list[float]:
        '''
        Returns the domain of the fuzzy variable.

        :return: list of floats.
        '''
        return self.linguistic_variables[0].domain


    def fuzzy_type(self) -> FUZZY_SETS:
        '''
        Returns the fuzzy type of the domain

        :return: the type of the fuzzy set present in the fuzzy variable.
        '''
        return self.fs_type


    def __getitem__(self, item) -> FS:
        '''
        Returns the corresponding fs.

        :param item: int. Index of the FS.
        :return: FS. The corresponding FS.
        '''
        return self.linguistic_variables[item]


    def __setitem__(self, item: int, elem: FS) -> None:
        '''
        Sets the corresponding fs.

        :param item: int. Index of the FS.
        :param elem: FS. The FS to set.
        '''
        self.linguistic_variables[item] = elem
    

    def __iter__(self) -> Generator[FS, None, None]:
        '''
        Returns the corresponding fs.

        :param item: int. Index of the FS.
        :return: FS. The corresponding FS.
        '''
        for fs in self.linguistic_variables:
            yield fs


    def __len__(self) -> int:
        '''
        Returns the number of linguistic variables.

        :return: int. Number of linguistic variables.
        '''
        return len(self.linguistic_variables)
    

    def __call__(self, x: np.array) -> list:
        '''
        Computes the membership to each of the FS in the fuzzy variables.

        :param x: numeric value or array.
        :return: list of floats. Membership to each of the FS in the fuzzy variables.
        '''
        return self.compute_memberships(x)





'''
Module that contains classifiers that uses two step genetic optimization and rule mining based on the support of the candidate rules. 

'''

import numpy as np
from sklearn.base import ClassifierMixin

try:
    from . import fuzzy_sets as fs
    from . import evolutionary_fit as evf
    from . import rule_mining as rm
    from . import utils
    from . import maintenance as mnt
except:
    import fuzzy_sets as fs
    import evolutionary_fit as evf
    import rule_mining as rm
    import utils
    import maintenance as mnt


class RuleMineClassifier(ClassifierMixin):
    """A classifier that works by mining a set of candidate rules with a minimum support, confidence and lift, and then using a genetic algorithm that chooses
    the optimal combination of those rules."""

    def __init__(self, nRules: int = 30, nAnts: int = 4, fuzzy_type: fs.FUZZY_SETS = None, tolerance: float = 0.0,
                verbose=False, n_class: int=None, runner: int=1, linguistic_variables: list[fs.fuzzyVariable]=None) -> None:
        '''
        Inits the optimizer with the corresponding parameters.

        :param nRules: number of rules to optimize.
        :param nAnts: max number of antecedents to use.
        :param fuzzy type: FUZZY_SET enum type in fuzzy_sets module. The kind of fuzzy set used.
        :param tolerance: tolerance for the support/dominance score of the rules.
        :param verbose: if True, prints the progress of the optimization.
        :param n_class: number of classes in the problem. If None (default) the classifier will compute it empirically.
        :param runner: number of threads to use.
        :param linguistic_variables: linguistic variables per antecedent.
        '''
        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.Classification]['data_mining'] += 1
        
        self.nAnts = nAnts
        self.fl_classifier = evf.BaseFuzzyRulesClassifier(nRules=nRules, linguistic_variables=linguistic_variables, 
                                            fuzzy_type=fuzzy_type, verbose=verbose, tolerance=tolerance, runner=runner, n_class=n_class)
        self.fuzzy_type = fuzzy_type
        self.tolerance = tolerance


    def fit(self, X: np.array, y: np.array, n_gen:int=30, pop_size:int=50, **kwargs) -> None:
        '''
        Trains the model with the given data.

        :param X: samples to train.
        :param y: labels for each sample.
        :param n_gen: number of generations to compute in the genetic optimization.
        :param pop_size: number of subjects per generation.
        :param kwargs: additional parameters for the genetic optimization. See fit method in BaseRuleBaseClassifier.
        '''
        fuzzy_vars = utils.construct_partitions(X, self.fuzzy_type)
        candidate_rules = rm.multiclass_mine_rulebase(X, y, fuzzy_vars, self.tolerance, max_depth=self.nAnts)
        self.fl_classifier.fit(X, y, checkpoints=0, candidate_rules=candidate_rules, n_gen=n_gen, pop_size=pop_size, **kwargs)


    def predict(self, X: np.array) -> np.array:
        '''
        Predict for each sample the corresponding class.

        :param X: samples to predict.
        :return: a class for each sample.
        '''
        # Make predictions using the fitted model
        y_pred = self.fl_classifier.predict(X)

        return y_pred
    

    def internal_classifier(self) -> evf.BaseFuzzyRulesClassifier:
        # Returns the classifier that performs the final predictions
        return self.fl_classifier
    


class FuzzyRulesClassifier(ClassifierMixin):
    """A classifier that works by performing a double optimization process. First, it creates a candidate rule base using genetic optimization
    and then uses it as a basis to obtain a better one that satisfies the constrain of antecedents and number of rules."""

    def __init__(self, nRules: int = 30, nAnts: int = 4, fuzzy_type: fs.FUZZY_SETS = None, tolerance: float = 0.0, 
                 verbose=False, n_class: int=None, runner: int=1, expansion_factor:int=1, linguistic_variables: list[fs.fuzzyVariable]=None) -> None:
        '''
        Inits the optimizer with the corresponding parameters.

        :param nRules: number of rules to optimize.
        :param nAnts: max number of antecedents to use.
        :param fuzzy type: FUZZY_SET enum type in fuzzy_sets module. The kind of fuzzy set used.
        :param tolerance: tolerance for the dominance score of the rules.
        :param verbose: if True, prints the progress of the optimization.
        :param n_class: number of classes in the problem. If None (default) the classifier will compute it empirically.
        :param runner: number of threads to use.
        :param expansion_factor: if > 1, it will compute inthe first optimization process n times the nRules parameters. (So that the search space for the second step is bigger)
        :param linguistic_variables: linguistic variables per antecedent.
        '''
        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.Classification]['double_go'] += 1


        self.fl_classifier1 = evf.BaseFuzzyRulesClassifier(nRules=nRules* expansion_factor, linguistic_variables=linguistic_variables, nAnts=nAnts, # We give this one more number rules so that then the second optimization has a bigger search space
                                            fuzzy_type=fuzzy_type, verbose=verbose, tolerance=tolerance, runner=runner, n_class=n_class) 
        self.fl_classifier2 = evf.BaseFuzzyRulesClassifier(nRules=nRules, linguistic_variables=linguistic_variables, nAnts=nAnts, 
                                            fuzzy_type=fuzzy_type, verbose=verbose, tolerance=tolerance, runner=runner, n_class=n_class)
        self.fuzzy_type = fuzzy_type
        self.tolerance = tolerance


    def fit(self, X: np.array, y: np.array, n_gen:int=30, pop_size:int=50, checkpoints:int=0, **kwargs) -> None:
        '''
        Trains the model with the given data.

        :param X: samples to train.
        :param y: labels for each sample.
        :param n_gen: number of generations to compute in the genetic optimization.
        :param pop_size: number of subjects per generation.
        :param checkpoints: if bigger than 0, will save the best subject per x generations in a text file.
        :param kwargs: additional parameters for the genetic optimization. See fit method in BaseRuleBaseClassifier.
        '''
        self.fl_classifier1.fit(X, y, n_gen, pop_size, checkpoints, **kwargs)
        self.phase1_rules = self.fl_classifier1.rule_base
        self.fl_classifier2.fit(X, y, n_gen, pop_size, checkpoints, initial_rules=self.phase1_rules, **kwargs)
        

    def predict(self, X: np.array) -> np.array:
        '''
        Predcit for each sample the correspondent class.

        :param X: samples to predict.
        :return: a class for each sample.
        '''
        # Make predictions using the fitted model
        y_pred = self.fl_classifier2.predict(X)

        return y_pred
    

    def internal_classifier(self) -> evf.BaseFuzzyRulesClassifier:
        # Returns the classifier that performs the final predictions
        return self.fl_classifier2
    

class RuleFineTuneClassifier(ClassifierMixin):
    """A classifier that works by mining a set of candidate rules with a minimum support and then uses a two step genetic optimization that chooses
    the optimal combination of those rules and fine tunes them."""

    def __init__(self, nRules: int = 30, nAnts: int = 4, fuzzy_type: fs.FUZZY_SETS = None, tolerance: float = 0.0, 
                 verbose=False, n_class: int=None, runner: int=1, expansion_factor:int=1, linguistic_variables: list[fs.fuzzyVariable]=None) -> None:
        '''
        Inits the optimizer with the corresponding parameters.

        :param nRules: number of rules to optimize.
        :param nAnts: max number of antecedents to use.
        :param fuzzy type: FUZZY_SET enum type in fuzzy_sets module. The kind of fuzzy set used.
        :param tolerance: tolerance for the dominance score of the rules.
        :param verbose: if True, prints the progress of the optimization.
        :param n_class: number of classes in the problem. If None (default) the classifier will compute it empirically.
        :param linguistic_variables: linguistic variables per antecedent.
        '''
        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.Classification]['double_go'] += 1
            mnt.usage_data[mnt.usage_categories.Classification]['data_mining'] += 1

        self.fl_classifier1 = evf.BaseFuzzyRulesClassifier(nRules=nRules* expansion_factor, linguistic_variables=linguistic_variables, nAnts=nAnts, # We give this one more number rules so that then the second optimization has a bigger search space
                                            fuzzy_type=fuzzy_type, verbose=verbose, tolerance=tolerance, runner=runner, n_class=n_class) 
        self.fl_classifier2 = evf.BaseFuzzyRulesClassifier(nRules=nRules, linguistic_variables=linguistic_variables, nAnts=nAnts, 
                                            fuzzy_type=fuzzy_type, verbose=verbose, tolerance=tolerance, runner=runner, n_class=n_class)
        self.fuzzy_type = fuzzy_type
        self.tolerance = tolerance


    def fit(self, X: np.array, y: np.array, n_gen:int=30, pop_size:int=50, checkpoints:int=0, **kwargs) -> None:
        '''
        Trains the model with the given data.

        :param X: samples to train.
        :param y: labels for each sample.
        :param n_gen: number of generations to compute in the genetic optimization.
        :param pop_size: number of subjects per generation.
        :param checkpoints: if bigger than 0, will save the best subject per x generations in a text file.
        :param kwargs: additional parameters for the genetic optimization. See fit method in BaseRuleBaseClassifier.
        '''
        candidate_rules = rm.multiclass_mine_rulebase(X, y, self.fl_classifier1.lvs, self.tolerance)

        self.fl_classifier1.fit(X, y, n_gen, pop_size, checkpoints, candidate_rules=candidate_rules, **kwargs)
        self.phase1_rules = self.fl_classifier1.rule_base
        self.fl_classifier2.fit(X, y, n_gen, pop_size, checkpoints, initial_rules=self.phase1_rules, **kwargs)


    def predict(self, X: np.array) -> np.array:
        '''
        Predcit for each sample the correspondent class.

        :param X: samples to predict.
        :return: a class for each sample.
        '''
        # Make predictions using the fitted model
        y_pred = self.fl_classifier.predict(X)

        return y_pred
    

    def internal_classifier(self) -> evf.BaseFuzzyRulesClassifier:
        # Returns the classifier that performs the final predictions
        return self.fl_classifier2


from . import centroid
from . import eval_rules
from . import eval_tools
from . import evolutionary_fit
from . import fuzzy_sets
from . import rules
from . import vis_rules
from . import utils
from . import persistence
from . import maintenance
from . import classifiers
